/*-----------------------------------------------------------------------
Copyright 2010-2016 S. GRAILLAT, F. JEZEQUEL
Copyright 2010 S. WANG, Y. ZHU
Copyright 2019-2020 J.E. TIDSWELL
Copyright 2023 F. JEZEQUEL

This file is part of SAM.

    SAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    SAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
    Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with SAM. If not, see
    <http://www.gnu.org/licenses/>.
-----------------------------------------------------------------------*/
#ifndef __SAM__
#define __SAM__

#include <iostream>
#include <stdio.h>
#include <cmath>
#include <time.h>
#include <stdint.h>
#include <string.h>

#include "gmp.h"
#include "mpfr.h"

#define SAM_NONE      0x00
#define SAM_DIV       0x1
#define SAM_MUL       0x2
#define SAM_POWER     0x4
#define SAM_MATH      0x8
#define SAM_BRANCHING 0x10
#define SAM_INTRINSIC 0x20
#define SAM_CANCEL    0x40
#define SAM_ALL       0xFFFFFFFF

#define LOG log10(2)
#define DEFAULT_PREC mpfr_get_default_prec()
#define DIGIT_NOT_COMPUTED -1


class mp_st
{

protected:
	mpfr_t x,y,z;
	typedef mpfr_t TYPEBASE;

private:
	mutable int accuracy;

public:

	// Constructors
	mp_st();
	mp_st(const mp_st& u); // copy
	mp_st(mp_st&& u); // move
	template<typename argT> mp_st(const argT& u);

	// destructor
	inline ~mp_st()
	{
		// by default mpfr_clear calls gmp_free
		// gmp_free checks for 0 size not null ptr
		if ((x->_mpfr_d != NULL)&&(y->_mpfr_d != NULL)&&(z->_mpfr_d != NULL)) {
			mpfr_clears(x, y, z, (mpfr_ptr)0 );
		}
	}


	// ASSIGNMENT

	mp_st& operator=(const mp_st&); // copy
	mp_st& operator=(mp_st&&); // move
	template<typename argT> mp_st& operator=(const argT& v);

	friend mp_st sam_set_str(mp_st &x, const char *s);
	//	friend void old_type(mp_st &b, mp_st &a); //FJ
	friend mp_st old_type(const mp_st &a); //FJ
	friend void SAM_prec_round(mp_st &a, int n); //FJ


	// ADDITION

	// unary +
	mp_st operator+() const { return *this; };
	// pre/post increment
	mp_st& operator++();
	mp_st operator++(int) { mp_st a(*this); ++(*this); return a; }

	mp_st& operator+=(const mp_st&);
	template<typename argT> mp_st& operator+=(const argT&);
	/* operator+(a,b) is defined below using a.operator+=(b) */


	// SUBTRACTION

	// unary -
	mp_st operator-() const ;
	// pre/post decrement
	mp_st& operator--();
	mp_st operator--(int) { mp_st a(*this); --(*this); return a; }

	mp_st& operator-=(const mp_st&);
	template<typename argT> mp_st& operator-=(const argT&);
	/* operator-(a,b) is defined below using a.operator-=(b) */


	// MULTIPLICATION

	mp_st& operator*=(const mp_st&);
	template<typename argT> mp_st& operator*=(const argT&);
	/* operator*(a,b) is defined below using a.operator*=(b) */


	// DIVISION

	mp_st& operator/=(const mp_st&);
	template<typename argT> mp_st& operator/=(const argT&);
	/* operator/(a,b) is defined below using a.operator/=(b) */


	// DEFINE an ORDER (used for branches and branch like relations)
	friend int order(const mp_st&, const mp_st&);
	friend int order(const mp_st&, const mpfr_t&);
	friend int order(const mpfr_t&, const mp_st&);
	template<typename argT> friend int order(const mp_st&, const argT&);
	template<typename argT> friend int order(const argT&, const mp_st&);


	// MATHEMATICAL FUNCTIONS

	friend mp_st  log(const mp_st&);
	friend mp_st  log2(const mp_st&);
	friend mp_st  log10(const mp_st&);
	friend mp_st  log1p(const mp_st&);
	friend mp_st  exp(const mp_st&);
	friend mp_st  exp2(const mp_st&);
	friend mp_st  expm1(const mp_st&);
	friend mp_st  sqrt(const mp_st&);
	friend mp_st  cbrt(const mp_st&);
	friend mp_st  sin(const mp_st&);
	friend mp_st  cos(const mp_st&);
	friend mp_st  tan(const mp_st&);
	friend mp_st  asin(const mp_st&);
	friend mp_st  acos(const mp_st&);
	friend mp_st  atan(const mp_st&);
	friend mp_st  atan2(const mp_st&, const mp_st&);
	friend mp_st  sinh(const mp_st&);
	friend mp_st  cosh(const mp_st&);
	friend mp_st  tanh(const mp_st&);
	friend mp_st  asinh(const mp_st&);
	friend mp_st  acosh(const mp_st&);
	friend mp_st  atanh(const mp_st&);
	friend mp_st  hypot(const mp_st&, const mp_st&);


	friend mp_st pow(const mp_st&, const mp_st&);
	template<typename argT> friend mp_st pow(const mp_st&, const argT&);
	template<typename argT> friend mp_st pow(const argT&, const mp_st&);


	// INTRINSIC FUNCTIONS

	friend mp_st fabs(mp_st);
	inline friend mp_st abs(const mp_st& a) { return fabs(a); }
	friend mp_st floor(mp_st);
	friend mp_st ceil(mp_st);
	friend mp_st trunc(mp_st);
	friend mp_st nearbyint(const mp_st&);
	friend mp_st rint(const mp_st&);
	friend long int  lrint(const mp_st&);
	friend long long int llrint(const mp_st&);

        // FMA

        friend mp_st fma(const mp_st&, const mp_st&, const mp_st&);

	// Conversion

	explicit operator int() const;
	explicit operator unsigned int() const;
	explicit operator long() const;
	explicit operator unsigned long() const;
	explicit operator float() const;
	explicit operator double() const;
	//	friend mpfr_t get_mpfr_t(const mp_st&); //FJ

	int nb_significant_digit() const;
	int computedzero() const;

	void display(const char *, const int prec, const int base=10) const ;
	void display(const char *s) const { display(s, DEFAULT_PREC*LOG, 10); }
	void display() const { display(NULL); }

	char* str( char *, size_t)  const ;
	friend char* strp(const mp_st&);
	friend char* str( char *, const mp_st&);

	friend std::istream& operator >>(std::istream& s, mp_st& );

	void data_st();
	void data_st(const double&, const int&);

	// class methods for special case coercion
	template<typename argT> 
	static mp_st coerce(const argT& x,const argT& y, const argT& z);

	// identical is intended for testing
	friend int identical(const mp_st&, const mp_st&);
	friend int identical(const mp_st&, const mpfr_t&);
	template<typename argT> friend int identical(const mp_st&, const argT&);
	template<typename argT> friend int identical(const argT& a, const mp_st& b) { return identical(b,a); }

	inline int getAccuracy() const {
		if (this->accuracy == DIGIT_NOT_COMPUTED) {
			this->accuracy = this->nb_significant_digit();
		}
		return this->accuracy;
	}
};


void  SAM_init(int, unsigned int, unsigned int, unsigned int,unsigned int);
void  SAM_init(int, unsigned int,unsigned int);
void  SAM_init(int, unsigned int);
void  SAM_end();

void self_validation_only();
void SAM_enable(unsigned int);
void SAM_disable(unsigned int);

std::ostream& operator<<(std::ostream&, const mp_st&);
std::istream& operator>>(std::istream&, mp_st& );


//
// operator+(a,b) exploiting move semantics
// (see https://stackoverflow.com/questions/16136461/move-semantics-and-operator-overloading)
//

template<typename argT> 
inline mp_st operator+(mp_st a, const argT& b) { a+=b; return a; }
template<typename argT> 
inline mp_st& operator+(mp_st&& a, const argT&& b) { a+=b; return a; }
template<typename argT> 
inline mp_st operator+(const argT&& a, mp_st b) { b+=a; return b; }
template<typename argT> 
inline mp_st& operator+(const argT& a, mp_st&& b) { b+=a; return a; }
// special case to resolve template ambiguity: both params as mp_st 
inline mp_st operator+(mp_st a, const mp_st& b) { a+=b; return a; }
inline mp_st& operator+(const mp_st& a, mp_st&& b) { b+=a; return b; }
inline mp_st& operator+(mp_st&& a, mp_st&& b) { a+=b; return a; }


//
// templated operator*(a,b) exploiting move semantics 
// (see https://stackoverflow.com/questions/16136461/move-semantics-and-operator-overloading)
//

template<typename argT> 
inline mp_st operator*(mp_st a, const argT& b) { a*=b; return a; }
template<typename argT> 
inline mp_st& operator*(mp_st&& a, const argT&& b) { a*=b; return a; }
template<typename argT> 
inline mp_st operator*(const argT&& a, mp_st b) { b*=a; return b; }
template<typename argT> 
inline mp_st& operator*(const argT& a, mp_st&& b) { b*=a; return b; }
// special case to resolve template ambiguity: both params as mp_st 
inline mp_st operator*(mp_st a, const mp_st& b) { a*=b; return a; }
inline mp_st& operator*(const mp_st& a, mp_st&& b) { b*=a; return b; }
inline mp_st& operator*(mp_st&& a, mp_st&& b) { a*=b; return a; }


//
// operator-(a,b) exploiting move semantics
// (see https://stackoverflow.com/questions/16136461/move-semantics-and-operator-overloading)
//

template<typename argT> 
inline mp_st operator-(mp_st a, const argT& b) { a-=b; return a; }
template<typename argT> 
inline mp_st& operator-(mp_st&& a, const argT&& b) { a-=b; return a; }
template<typename argT> 
inline mp_st operator-(const argT&& a, mp_st b) { mp_st res(a); res-=b; return res; }
template<typename argT> 
inline mp_st& operator-(const argT& a, mp_st&& b) {  b-=a; b=-b; return b; } // TODO : implement -() move semantics
// special case to resolve template ambiguity: both params as mp_st 
inline mp_st operator-(mp_st a, const mp_st& b) { a-=b; return a; }
inline mp_st& operator-(const mp_st& a, mp_st&& b) { b-=a; b=-b; return b; } // TODO : implement -() move semantics
inline mp_st& operator-(mp_st&& a, mp_st&& b) { a-=b; return a; }


//
// operator/(a,b) exploiting move semantics
// (see https://stackoverflow.com/questions/16136461/move-semantics-and-operator-overloading)
//

template<typename argT> 
inline mp_st operator/(mp_st a, const argT& b) { a/=b; return a; }
template<typename argT> 
inline mp_st& operator/(mp_st&& a, const argT&& b) { a/=b; return a; }
template<typename argT> 
inline mp_st operator/(const argT& a, const mp_st&& b) { mp_st res(a); res/=b; return res;  }


// relational operators defined using order()
#include "SAM_relations.hpp"



#endif // __SAM__











