/*-----------------------------------------------------------------------
Copyright 2020 J Tidswell

This file is part of SAM.

    SAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    SAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
    Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with SAM. If not, see
    <http://www.gnu.org/licenses/>.
-----------------------------------------------------------------------*/
///////////////////////////////////////////////////
//
// Provides simple template wrapper for common MPFR functions to make code appear more generic
//
///////////////////////////////////////////////////

#ifndef __SAM_MPFR_TEMPLATES__
#define __SAM_MPFR_TEMPLATES__


#include "mpfr.h"
#include <cassert>

// use a private namespace so that the MPFR definitions specifically for SAM dont clash with other projects
namespace SAM {



/* stack allocate small mpfr */
/* modified version of the official macro */
/*
* #define MPFR_DECL_INIT(_x, _p)                                        \
*		MPFR_EXTENSION mp_limb_t __gmpfr_local_tab_##_x[((_p)-1)/GMP_NUMB_BITS+1]; \
*		MPFR_EXTENSION mpfr_t _x = {{(_p),1,__MPFR_EXP_NAN,__gmpfr_local_tab_##_x}}
*/
inline void mpfr_init_local(mpfr_t &x, mp_limb_t* limb )
{
	const unsigned short float_signficand_size = 24;
	assert(GMP_NUMB_BITS>float_signficand_size); /*default is 32 of 64 so this is reasonable */
	x[0]._mpfr_prec = float_signficand_size;
	x[0]._mpfr_sign = 1;
	x[0]._mpfr_exp = __MPFR_EXP_NAN;             
	x[0]._mpfr_d = limb;
}
inline void mpfr_clear_local(mpfr_t &x) 
{
	// useful for error catching, disable in release builds ?
	x->_mpfr_d = NULL;
	x->_mpfr_exp = __MPFR_EXP_NAN;
	x->_mpfr_prec = 0;
	x->_mpfr_sign = 0;
}


/* borrowing more from MPFR */
// MPFR does not have a kind function for general init variables ...
inline mpfr_kind_t mpfr_get_kind (const mpfr_t& x)
{
	const mpfr_exp_t k = x->_mpfr_exp;
	mpfr_kind_t kind;
	if ( k == __MPFR_EXP_NAN ) { kind = MPFR_NAN_KIND; }
	else if ( k == __MPFR_EXP_INF ) { kind = MPFR_INF_KIND; }
	else if ( k == __MPFR_EXP_ZERO ) { kind = MPFR_ZERO_KIND; }
	else /* k > __MPFR_EXP_INF */ { kind = MPFR_REGULAR_KIND; }
	return kind;
}



//
//  Addition
//

// operator +(src,a)
inline int add_mpfr( mpfr_ptr res, mpfr_srcptr src, const unsigned int& a, mpfr_rnd_t dirn ) { return mpfr_add_ui( res, src, a, dirn ); }
inline int add_mpfr( mpfr_ptr res, mpfr_srcptr src, const int& a, mpfr_rnd_t dirn ) { return mpfr_add_si( res, src, a, dirn ); }
inline int add_mpfr( mpfr_ptr res, mpfr_srcptr src, const unsigned long& a, mpfr_rnd_t dirn ) { return mpfr_add_ui( res, src, a, dirn ); }
inline int add_mpfr( mpfr_ptr res, mpfr_srcptr src, const long& a, mpfr_rnd_t dirn ) { return mpfr_add_si( res, src, a, dirn ); }
inline int add_mpfr( mpfr_ptr res, mpfr_srcptr src, const double& a, mpfr_rnd_t dirn ) { return mpfr_add_d( res, src, a, dirn ); }
inline int add_mpfr( mpfr_ptr res, mpfr_srcptr src, const float& a, mpfr_rnd_t dirn ) { return mpfr_add_d( res, src, (double)a, dirn ); }
inline int add_mpfr( mpfr_ptr res, mpfr_srcptr src, const mpfr_srcptr& a, mpfr_rnd_t dirn ) { return mpfr_add( res, src, a, dirn ); }

template<typename T>
inline int add_mpfr( mpfr_ptr res, const T& x, mpfr_srcptr a, mpfr_rnd_t dirn ) { return add_mpfr( res, a, x, dirn ); }

// operator +=(a)
template<typename T>
inline int add_mpfr( mpfr_ptr x, const T& a, mpfr_rnd_t dirn ) { return add_mpfr( x, (mpfr_srcptr)x, a, dirn ); }


//
//  Subtraction
//

// operator -(src,a)
inline int sub_mpfr( mpfr_ptr res, mpfr_srcptr src, const unsigned int& a, mpfr_rnd_t dirn ) { return mpfr_sub_ui( res, src, a, dirn ); }
inline int sub_mpfr( mpfr_ptr res, mpfr_srcptr src, const int& a, mpfr_rnd_t dirn ) { return mpfr_sub_si( res, src, a, dirn ); }
inline int sub_mpfr( mpfr_ptr res, mpfr_srcptr src, const unsigned long& a, mpfr_rnd_t dirn ) { return mpfr_sub_ui( res, src, a, dirn ); }
inline int sub_mpfr( mpfr_ptr res, mpfr_srcptr src, const long& a, mpfr_rnd_t dirn ) { return mpfr_sub_si( res, src, a, dirn ); }
inline int sub_mpfr( mpfr_ptr res, mpfr_srcptr src, const double& a, mpfr_rnd_t dirn ) { return mpfr_sub_d( res, src, a, dirn ); }
inline int sub_mpfr( mpfr_ptr res, mpfr_srcptr src, const float& a, mpfr_rnd_t dirn ) { return mpfr_sub_d( res, src, (double)a, dirn ); }
inline int sub_mpfr( mpfr_ptr res, mpfr_srcptr src, const mpfr_srcptr& a, mpfr_rnd_t dirn ) { return mpfr_sub( res, src, a, dirn ); }

inline int sub_mpfr( mpfr_ptr res, const unsigned int& a, mpfr_srcptr src, mpfr_rnd_t dirn ) { return mpfr_ui_sub( res, a, src, dirn ); }
inline int sub_mpfr( mpfr_ptr res, const int& a, mpfr_srcptr src, mpfr_rnd_t dirn ) { return mpfr_si_sub( res, a, src, dirn ); }
inline int sub_mpfr( mpfr_ptr res, const unsigned long& a, mpfr_srcptr src, mpfr_rnd_t dirn ) { return mpfr_ui_sub( res, a, src, dirn ); }
inline int sub_mpfr( mpfr_ptr res, const long& a, mpfr_srcptr src, mpfr_rnd_t dirn ) { return mpfr_si_sub( res, a, src, dirn ); }
inline int sub_mpfr( mpfr_ptr res, const double& a, mpfr_srcptr src, mpfr_rnd_t dirn ) { return mpfr_d_sub( res, a, src, dirn ); }
inline int sub_mpfr( mpfr_ptr res, const float& a, mpfr_srcptr src, mpfr_rnd_t dirn ) { return mpfr_d_sub( res, (double)a, src, dirn ); }

// operator -=(a)
template<typename T>
inline int sub_mpfr( mpfr_ptr x, const T& a, mpfr_rnd_t dirn ) { return sub_mpfr( x, (mpfr_srcptr)x, a, dirn ); }


//
//  Multiplication
//

// operator *(src,a)
inline int mul_mpfr( mpfr_ptr res, mpfr_srcptr src, const unsigned long& a, mpfr_rnd_t dirn ) { return mpfr_mul_ui( res, src, a, dirn ); }
inline int mul_mpfr( mpfr_ptr res, mpfr_srcptr src, const long& a, mpfr_rnd_t dirn ) { return mpfr_mul_si( res, src, a, dirn ); }
inline int mul_mpfr( mpfr_ptr res, mpfr_srcptr src, const unsigned int& a, mpfr_rnd_t dirn ) { return mpfr_mul_ui( res, src, a, dirn ); }
inline int mul_mpfr( mpfr_ptr res, mpfr_srcptr src, const int& a, mpfr_rnd_t dirn ) { return mpfr_mul_si( res, src, a, dirn ); }
inline int mul_mpfr( mpfr_ptr res, mpfr_srcptr src, const double& a, mpfr_rnd_t dirn ) { return mpfr_mul_d( res, src, a, dirn ); }
inline int mul_mpfr( mpfr_ptr res, mpfr_srcptr src, const float& a, mpfr_rnd_t dirn ) { return mpfr_mul_d( res, src, (double)a, dirn ); }
inline int mul_mpfr( mpfr_ptr res, mpfr_srcptr src, const mpfr_srcptr& a, mpfr_rnd_t dirn ) { return mpfr_mul( res, src, a, dirn ); }

template<typename T>
inline int mul_mpfr( mpfr_ptr res, const T&  src, const mpfr_srcptr& a, mpfr_rnd_t dirn ) { return mul_mpfr( res, a, src, dirn ); }

// operator *=(a)
template<typename T>
inline int mul_mpfr( mpfr_ptr x, const T& a, mpfr_rnd_t dirn ) { return mul_mpfr( x, (mpfr_srcptr)x, a, dirn ); }


//
//  Division
//

// operator /(src,a)
inline int div_mpfr( mpfr_ptr res, mpfr_srcptr src, const unsigned long& a, mpfr_rnd_t dirn ) { return mpfr_div_ui( res, src, a, dirn ); }
inline int div_mpfr( mpfr_ptr res, mpfr_srcptr src, const long& a, mpfr_rnd_t dirn ) { return mpfr_div_si( res, src, a, dirn ); }
inline int div_mpfr( mpfr_ptr res, mpfr_srcptr src, const unsigned int& a, mpfr_rnd_t dirn ) { return mpfr_div_ui( res, src, a, dirn ); }
inline int div_mpfr( mpfr_ptr res, mpfr_srcptr src, const int& a, mpfr_rnd_t dirn ) { return mpfr_div_si( res, src, a, dirn ); }
inline int div_mpfr( mpfr_ptr res, mpfr_srcptr src, const double& a, mpfr_rnd_t dirn ) { return mpfr_div_d( res, src, a, dirn ); }
inline int div_mpfr( mpfr_ptr res, mpfr_srcptr src, const float& a, mpfr_rnd_t dirn ) { return mpfr_div_d( res, src, (double)a, dirn ); }
inline int div_mpfr( mpfr_ptr res, mpfr_srcptr src, const mpfr_srcptr& a, mpfr_rnd_t dirn ) { return mpfr_div( res, src, a, dirn ); }

inline int div_mpfr( mpfr_ptr res, const unsigned int& a, mpfr_srcptr src, mpfr_rnd_t dirn ) { return mpfr_ui_div( res, a, src, dirn ); }
inline int div_mpfr( mpfr_ptr res, const int& a, mpfr_srcptr src, mpfr_rnd_t dirn ) { return mpfr_si_div( res, a, src, dirn ); }
inline int div_mpfr( mpfr_ptr res, const unsigned long& a, mpfr_srcptr src, mpfr_rnd_t dirn ) { return mpfr_ui_div( res, a, src, dirn ); }
inline int div_mpfr( mpfr_ptr res, const long& a, mpfr_srcptr src, mpfr_rnd_t dirn ) { return mpfr_si_div( res, a, src, dirn ); }
inline int div_mpfr( mpfr_ptr res, const double& a, mpfr_srcptr src, mpfr_rnd_t dirn ) { return mpfr_d_div( res, a, src, dirn ); }
inline int div_mpfr( mpfr_ptr res, const float& a, mpfr_srcptr src, mpfr_rnd_t dirn ) { return mpfr_d_div( res, (double)a, src, dirn ); }

// operator /=(a)
template<typename T>
inline int div_mpfr( mpfr_ptr x, const T& a, mpfr_rnd_t dirn ) { return div_mpfr( x, (mpfr_srcptr)x, a, dirn ); }


//
//  Construction
//

inline int set_mpfr( mpfr_ptr x, const unsigned long& a, mpfr_rnd_t dirn ) { return mpfr_set_ui( x, a, dirn ); }
inline int set_mpfr( mpfr_ptr x, const long& a, mpfr_rnd_t dirn ) { return mpfr_set_si( x, a, dirn ); }
inline int set_mpfr( mpfr_ptr x, const unsigned int& a, mpfr_rnd_t dirn ) { return mpfr_set_ui( x, a, dirn ); }
inline int set_mpfr( mpfr_ptr x, const int& a, mpfr_rnd_t dirn ) { return mpfr_set_si( x, a, dirn ); }
inline int set_mpfr( mpfr_ptr x, const double& a, mpfr_rnd_t dirn ) { return mpfr_set_d( x, a, dirn ); }
inline int set_mpfr( mpfr_ptr x, const float& a, mpfr_rnd_t dirn ) { return mpfr_set_flt( x, a, dirn ); }
inline int set_mpfr( mpfr_ptr x, const mpfr_srcptr& a, mpfr_rnd_t dirn ) { return mpfr_set( x, a, dirn ); }


//
//  power
//

inline int pow_mpfr( mpfr_ptr res, mpfr_srcptr b, const unsigned long& a, mpfr_rnd_t dirn ) { return mpfr_pow_ui( res, b, a, dirn ); }
inline int pow_mpfr( mpfr_ptr res, mpfr_srcptr b, const long& a, mpfr_rnd_t dirn ) { return mpfr_pow_si( res, b, a, dirn ); }
inline int pow_mpfr( mpfr_ptr res, mpfr_srcptr b, const unsigned int& a, mpfr_rnd_t dirn ) { return mpfr_pow_ui( res, b, a, dirn ); }
inline int pow_mpfr( mpfr_ptr res, mpfr_srcptr b, const int& a, mpfr_rnd_t dirn ) { return mpfr_pow_si( res, b, a, dirn ); }
inline int pow_mpfr( mpfr_ptr res, mpfr_srcptr b, const mpfr_srcptr& a, mpfr_rnd_t dirn ) { return mpfr_pow( res, b, a, dirn ); }

inline int pow_mpfr( mpfr_ptr res, unsigned long b, const mpfr_srcptr& a, mpfr_rnd_t dirn ) { return mpfr_ui_pow( res, b, a, dirn ); }
inline int pow_mpfr( mpfr_ptr res, unsigned long b, const unsigned long& a, mpfr_rnd_t dirn ) { return mpfr_ui_pow_ui( res, a, b, dirn ); }
inline int pow_mpfr( mpfr_ptr res, unsigned long b, const unsigned int& a, mpfr_rnd_t dirn ) { return mpfr_ui_pow_ui( res, a, b, dirn ); }


//
//  Comparison
//

// a <?> b
inline int cmp_mpfr( const mpfr_srcptr& a, const mpfr_srcptr& b ) { return mpfr_cmp( a, b ); };
inline int cmp_mpfr( const mpfr_srcptr& a, const unsigned int& b ) { return mpfr_cmp_ui( a, b ); }
inline int cmp_mpfr( const mpfr_srcptr& a, const int& b ) { return mpfr_cmp_si( a, b ); }
inline int cmp_mpfr( const mpfr_srcptr& a, const unsigned long& b ) { return mpfr_cmp_ui( a, b ); }
inline int cmp_mpfr( const mpfr_srcptr& a, const long& b ) { return mpfr_cmp_si( a, b ); }
inline int cmp_mpfr( const mpfr_srcptr& a, const double& b ) { return mpfr_cmp_d( a, b ); }
inline int cmp_mpfr( const mpfr_srcptr& a, const float& b ) { return mpfr_cmp_d( a, b ); }
template<typename T>
inline int cmp_mpfr( const T& a, const mpfr_srcptr& b ) { return -cmp_mpfr(b,a); };


//
//  Inequalities
//

// a >= b
inline int cmp_mpfr_ge( const mpfr_srcptr& a, const mpfr_srcptr& b ) { return mpfr_greaterequal_p(a,b); }
template<typename Ta, typename Tb> 
inline int cmp_mpfr_ge( const Ta& a, const Tb& b ) { return (cmp_mpfr(a,b) >= 0); }

// a > b
inline int cmp_mpfr_gt( const mpfr_srcptr& a, const mpfr_srcptr& b ) { return mpfr_greater_p(a,b); }
template<typename Ta, typename Tb> 
inline int cmp_mpfr_gt( Ta& a, const Tb& b ) { return (cmp_mpfr(a,b) > 0); }

// a <= b
inline int cmp_mpfr_le( const mpfr_srcptr& a, const mpfr_srcptr& b ) { return mpfr_lessequal_p(a,b); }
template<typename Ta, typename Tb> 
inline int cmp_mpfr_le( const Ta& a, const Tb& b ) { return (cmp_mpfr(a,b) <= 0); }

// a < b
inline int cmp_mpfr_lt( const mpfr_srcptr& a, const mpfr_srcptr& b ) { return mpfr_less_p(a,b); }
template<typename Ta, typename Tb> 
inline int cmp_mpfr_lt( const Ta& a, const Tb& b ) { return (cmp_mpfr(a,b) < 0); }

} // SAM namespace


#endif // __SAM_MPFR_TEMPLATES__
