/*-----------------------------------------------------------------------
Copyright 2010-2016 S. GRAILLAT, F. JEZEQUEL
Copyright 2010 S. WANG, Y. ZHU
Copyright 2019-2020 J.E. TIDSWELL

This file is part of SAM.

    SAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    SAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
    Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with SAM. If not, see
    <http://www.gnu.org/licenses/>.
-----------------------------------------------------------------------*/
///////////////////////////////////////////////////
//
// SAM SOFTWARE
//
///////////////////////////////////////////////////
// Last Modifications
//
//
///////////////////////////////////////////////////

#include <cassert>
#include <cmath>

#include "SAM.h"
#include "SAM_private.h"
#include "SAM_MPFR_templates.h"

using namespace SAM;

// ----------------------------------------


// save the rounding mode to compute the number of significant digits
// always with the same rounding mode


//****f* SAM_digitnumber/digitnumber
//    NAME
//      digitnumber
//    SYNOPSIS
//      res = digitnumber(x)
//      res = x.digitnumber()
//    FUNCTION
//      The digitnumber() function returns the number of significant
//      digits of a stochastic x
//
//
//    INPUTS
//      x           - a stochastic number
//    RESULT
//      res         - an integer value
//
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//
//
//    BUGS
//
//
//    SEE ALSO
//
//
//  /

using namespace SAM; // expose the SAM specific MPFR code

int mp_st::nb_significant_digit() const
{
	const mpfr_prec_t default_prec = DEFAULT_PREC;
	const long MAX_PREC = (long)((default_prec*LOG));
	
	const mpfr_kind_t xkind = mpfr_get_kind (x);
	const mpfr_kind_t ykind = mpfr_get_kind (y);
	const mpfr_kind_t zkind = mpfr_get_kind (z);

	if ( (xkind == ykind) && (xkind == zkind) )
	{
		// all NaNs is stochastic zero
		if (xkind == MPFR_NAN_KIND ) { accuracy = 0; return accuracy; }
		// all Zero is exact zero
		if (xkind == MPFR_ZERO_KIND ) { accuracy = MAX_PREC; return accuracy; }

		const mpfr_sign_t xsign = mpfr_sgn(x);
		const mpfr_sign_t ysign = mpfr_sgn(y);
		const mpfr_sign_t zsign = mpfr_sgn(z);
		// mixed signs is a stochastic zero (for Infinity or Regular)
		if ( (xsign != ysign) || (xsign != zsign) ) { accuracy = 0; return accuracy; }
#if defined(JT_SEMANTICS) && (JT_SEMANTICS>0)
		// all Infinite with same sign is exact
		if ( xkind == MPFR_INF_KIND ) { accuracy = MAX_PREC; return accuracy; }
#else // PEQUAN_SEMANTICS
		// Infinity is probably overflow, which is loss of representation, so its a stochastic zero
		if ( xkind == MPFR_INF_KIND ) { accuracy = 0; return accuracy; }
#endif
		/* this excludes many special values so that following code is simpler and cleaner */
		/* including it excludes anything with a mean of 0, exact or otherwise */

		float xf, d2f, d3f, muf, vf;
		long xe, d2e, d3e, dde;
		float lgf;
		int ve, mue;

		/* reduced precision can be stack allocated */
		mpfr_t d2, d3;
		mp_limb_t d2_l, d3_l;
		mpfr_init_local(d2, &d2_l); /* do not need to clear local init */
		mpfr_init_local(d3, &d3_l);
		
		/* dont exit suddenly between changing and restoring mpfr state */
		// mp_rnd_t previous_rnd = mpfr_get_default_rounding_mode(); // save mpfr state
		// mpfr_set_default_rounding_mode(MPFR_RNDN);                // change mpfr state
		// mp_rnd_t tmp_rnd = mpfr_get_default_rounding_mode();

		/* calculate differences from full precision */
		mpfr_sub( d2, x, y, MPFR_RNDN );
		mpfr_sub( d3, x, z, MPFR_RNDN );

		d2f = (float)mpfr_get_d_2exp( &d2e, d2, MPFR_RNDN );
		d3f = (float)mpfr_get_d_2exp( &d3e, d3, MPFR_RNDN );
		xf = (float)mpfr_get_d_2exp( &xe, x, MPFR_RNDN );

		// mpfr_set_default_rounding_mode(previous_rnd);             // restore mpfr state

		// zero differences means exact 
		/* this excludes zero variance, so no division by zero of log of zero errors */
		if ( (d2f == 0) && (d3f == 0) )
		{
			accuracy = MAX_PREC;
			return accuracy;
		}

		/* normalisation to first sample */
		/* cancelled by calculation of error ratio */
		d2e -= xe;
		d3e -= xe;

		/* normalisation of differences */
		if ( d2f == 0 ) { dde = d3e; }
		else if ( d3f == 0 ) { dde = d2e; }
		else { dde = std::max(d2e,d3e); }
		d2e -= dde;
		d3e -= dde;
		d2f = scalbnf(d2f,d2e);
		d3f = scalbnf(d3f,d3e);

		/* calculate mean - adjust differences to normalised mean */
		muf = d2f + d3f;
		muf *= (float)(1.0/3.0);
		muf = scalbnf(muf,dde);
		muf += xf;
		/* logarithm approximations assume significand in [1,2) */
		muf = frexpf(muf,&mue);
		/* frexpf() returns signifiand in [0.5,1) */
		muf *= 2; mue -= 1;

		/* calculate variance */
		vf = d2f * d2f;
		vf -= d2f * d3f;
		vf += d3f * d3f;
		// vf *= (float)(1.0/3.0);
		// vf *= 3.0854661704166664; // include magic constant ! (before frexp)
		vf *= 1.0284887234722222f; // fold constants together
		/* logarithm approximations assume significand in [1,2) */
		vf = frexpf(vf, &ve);
		/* frexpf() returns signifiand in [0.5,1) */
		vf *= 2; ve -= 1;
		/* variance is square of difference terms */
		// ve += dde * 2; // dont double then halve in log, it helps avoid overflow

		/* new approx log */
		/*   
		*  log10(abs(mean)) - 0.5*log10(variance)
		*  log10(2)(log2(abs(mean)) - 0.5*log2(variance))
		*  understate log of mean overstate log of variance
		*/
		// lgf = (abs(muf)-1) + mue; // log(abs(mean)) base 2
		// lgf -= 0.5 * ((vf-1)*(3-vf) + ve); // log(variance) base 2
		// careful to avoid overflow !
		lgf = ( abs(muf)-1 ) + mue ; // approx log(abs(mean)) base 2
		lgf -= 0.5 * ( (vf-1)*(3-vf) + (float)ve ); // approx log(abs(variance)) base 2
		lgf -= dde;  // we skipped adding this in variance
		lgf *= LOG; // convert to base 10
		accuracy = (int)lgf;

		/* clamp accuracy in bounds */
		accuracy = ( accuracy > MAX_PREC ) ? MAX_PREC : accuracy ;
		accuracy = ( accuracy < 0 ) ? 0 : accuracy ;
	} else {  // kind not all the same
		// mix of zero, infinity, nan with anything (including regular) has to be stochastic zero
		// more complicated if kind includes SUBNORMAL but MPFR does not 
		accuracy = 0;
	}

	return accuracy;
}


