/*-----------------------------------------------------------------------
Copyright 2010-2016 S. GRAILLAT, F. JEZEQUEL
Copyright 2010 S. WANG, Y. ZHU
Copyright 2019-2020 J.E. TIDSWELL

This file is part of SAM.

	SAM is free software: you can redistribute it and/or modify it
	under the terms of the GNU Lesser General Public License as
	published by the Free Software Foundation, either version 3 of the
	License, or (at your option) any later version.

	SAM is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
	or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
	Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with SAM. If not, see
	<http://www.gnu.org/licenses/>.
-----------------------------------------------------------------------*/
/////////////////////////////////////////////////////
// SAM SOFTWARE
//
///////////////////////////////////////////////////
// Last Modifications
//
//
///////////////////////////////////////////////////


#include "SAM.h"
#include "SAM_private.h"
#include "SAM_MPFR_templates.h"

using namespace SAM;


//****f* SAM_math/pow
//    NAME
//      pow
//    SYNOPSIS
//      res = pow(x,y)
//    FUNCTION
//      The pow() function computes x raised to the power y.
//
//
//    INPUTS
//      a           - mp_st
//      b           - mp_st
//    RESULT
//      res         - mp_st
//
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//
//
//    BUGS
//
//
//    SEE ALSO
//
//
//  /




/////////////////////////////////////////


mp_st pow(const mp_st& a, const mp_st& b)
{
	mp_st res;
	if(_SAM_power_tag!=0){
		if (a.accuracy==DIGIT_NOT_COMPUTED) a.nb_significant_digit();
		if (b.accuracy==DIGIT_NOT_COMPUTED) b.nb_significant_digit();
		if (a.accuracy==0 || b.accuracy==0) instability(&_SAM_power_count);
	}
	res.accuracy=DIGIT_NOT_COMPUTED;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	SAM::pow_mpfr( res.x, a.x, b.x, xdir );
	SAM::pow_mpfr( res.y, a.y, b.y, ydir );
	SAM::pow_mpfr( res.z, a.z, b.z, zdir );

	return res;
}


template<typename argT> mp_st pow(const mp_st& a, const argT& b)
{
	mp_st res;
	if(_SAM_power_tag!=0){
		if (a.accuracy==DIGIT_NOT_COMPUTED) a.nb_significant_digit();
		if (a.accuracy==0) instability(&_SAM_power_count);
	}
	res.accuracy=DIGIT_NOT_COMPUTED;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	SAM::pow_mpfr( res.x, a.x, b, xdir );
	SAM::pow_mpfr( res.y, a.y, b, ydir );
	SAM::pow_mpfr( res.z, a.z, b, zdir );

	return res;
}

template<typename argT> mp_st pow(const argT& a, const mp_st& b)
{
	mp_st res;
	if(_SAM_power_tag!=0){
		if (b.accuracy==DIGIT_NOT_COMPUTED) b.nb_significant_digit();
		if (b.accuracy==0) instability(&_SAM_power_count);
	}
	res.accuracy=DIGIT_NOT_COMPUTED;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	SAM::pow_mpfr( res.x, a, b.x, xdir );
	SAM::pow_mpfr( res.y, a, b.y, ydir );
	SAM::pow_mpfr( res.z, a, b.z, zdir );

	return res;
}


// need to allocate a temporary for types mpfr does not implement directly
template<typename pT>
static mp_st apow(const mp_st& b, const pT& a)
{
	mpfr_t atmp;
	mpfr_init2 (atmp, DEFAULT_PREC);
	SAM::set_mpfr (atmp, a, MPFR_RNDN);

	mp_st res = pow(b, atmp);

	mpfr_clear(atmp);
	return res;
}

// need to allocate a temporary for types mpfr does not implement directly
template<typename pT>
static mp_st bpow(const pT& b, const mp_st& a)
{
	mpfr_t btmp;
	mpfr_init2 (btmp, DEFAULT_PREC);
	SAM::set_mpfr (btmp, b, MPFR_RNDN);

	mp_st res = pow(btmp, a);

	mpfr_clear(btmp);
	return res;
}

//
// explicit instantiation so compiler generates code to include in linker
//

template mp_st pow(const mp_st&, const int&);
template mp_st pow(const mp_st&, const unsigned int&);
template mp_st pow(const mp_st&, const long&);
template mp_st pow(const mp_st&, const unsigned long&);
template mp_st pow(const mp_st&, const mpfr_t&);
template mp_st pow(const mp_st&, const mpfr_srcptr&);
template<> mp_st pow<double>(const mp_st& b, const double& a) { return apow(b, a); }
template<> mp_st pow<float>(const mp_st& b, const float& a) { return apow(b, a); }

template mp_st pow(const unsigned int&, const mp_st&);
template mp_st pow(const unsigned long&, const mp_st&);
template mp_st pow(const mpfr_t&, const mp_st&);
template mp_st pow(const mpfr_srcptr&, const mp_st&);
template<> mp_st pow<int>(const int& b, const mp_st& a) { return bpow(b, a); }
template<> mp_st pow<long>(const long& b, const mp_st& a) { return bpow(b, a); }
template<> mp_st pow<float>(const float& b, const mp_st& a) { return bpow(b, a); }
template<> mp_st pow<double>(const double& b, const mp_st& a) { return bpow(b, a); }



//////////////////////////////////////////////////////////////////////////////

//****f* SAM_math/log
//    NAME
//      log
//
//    SYNOPSIS
//      res = log(x)
//
//    FUNCTION
//       The log() function computes the value of the natural
//       logarithm of argument x.
//
//    INPUTS
//      a           - mp_st
//    RESULT
//      res         - mp_st
//    SEE ALSO
//      log2() , log10() , log1p(), exp(3), exp2(3), expm1(3), pow(3)
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//
//
//    BUGS
//
//


mp_st log(const mp_st& a)
{
	mp_st res;
	if(_SAM_math_tag!=0 || mpfr_sgn(a.x)<0){
		if (a.accuracy==DIGIT_NOT_COMPUTED) a.nb_significant_digit();
		if (a.accuracy==0) instability(&_SAM_math_count);
	}
	res.accuracy=DIGIT_NOT_COMPUTED;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	mpfr_log( res.x, a.x, xdir );
	mpfr_log( res.y, a.y, ydir );
	mpfr_log( res.z, a.z, zdir );

	return res;
}

//////////////////////////////////////////////////////////////////////////////

//****f* SAM_math/log2
//    NAME
//      log2
//    SYNOPSIS
//      res = log2(x)
//    FUNCTION
//      The log2() function computes the value of the logarithm of
//      argument x to base 2.
//
//
//    INPUTS
//      a           - mp_st
//    RESULT
//      res         - mp_st
//    SEE ALSO
//      log() , log10() , log1p(), exp(3), exp2(3), expm1(3), pow(3)
//
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//
//
//    BUGS
//
//
//    SEE ALSO
//
//
//  /


mp_st log2(const mp_st& a)
{
	mp_st res;
	if(_SAM_math_tag!=0 || mpfr_sgn(a.x)<0){
		if (a.accuracy==DIGIT_NOT_COMPUTED) a.nb_significant_digit();
		if (a.accuracy==0) instability(&_SAM_math_count);
	}
	res.accuracy=DIGIT_NOT_COMPUTED;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	mpfr_log2( res.x, a.x, xdir );
	mpfr_log2( res.y, a.y, ydir );
	mpfr_log2( res.z, a.z, zdir );

	return res;
}

//////////////////////////////////////////////////////////////////////////////

//****f* SAM_math/log10
//    NAME
//       log10
//    SYNOPSIS
//      res =  log10(x)
//    FUNCTION
//      The log10() function computes the value of the logarithm of
//      argument x to base 10.
//
//
//    INPUTS
//      a           - mp_st
//    RESULT
//      res         - mp_st
//    SEE ALSO
//      log() , log2() , log1p(), exp(3), exp2(3), expm1(3), pow(3)
//
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//
//
//    BUGS
//
//
//    SEE ALSO
//
//
//  /


mp_st log10(const mp_st& a)
{
	mp_st res;
	if(_SAM_math_tag!=0 || mpfr_sgn(a.x)<0){
		if (a.accuracy==DIGIT_NOT_COMPUTED) a.nb_significant_digit();
		if (a.accuracy==0) instability(&_SAM_math_count);
	}
	res.accuracy=DIGIT_NOT_COMPUTED;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	mpfr_log10( res.x, a.x, xdir );
	mpfr_log10( res.y, a.y, ydir );
	mpfr_log10( res.z, a.z, zdir );
	return res;
}

//////////////////////////////////////////////////////////////////////////////

//****f* SAM_math/log1p
//    NAME
//       log1p
//    SYNOPSIS
//      res =  log1p(x)
//    FUNCTION
//      The log1p() function computes the value of log(1+x) accurately
//      even for very small values of x.
//
//
//    INPUTS
//      a           - mp_st
//    RESULT
//      res         - mp_st
//    SEE ALSO
//      log() , log2() , log10(), exp(3), exp2(3), expm1(3), pow(3)
//
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//
//
//    BUGS
//
//
//    SEE ALSO
//
//
//  /


mp_st log1p(const mp_st& a)
{
	mp_st res;
	if(_SAM_math_tag!=0 || mpfr_cmp_ui(a.x, -1)<=0){
		if (a.accuracy==DIGIT_NOT_COMPUTED) a.nb_significant_digit();
		if (a.accuracy==0) instability(&_SAM_math_count);
	}
	res.accuracy=DIGIT_NOT_COMPUTED;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	mpfr_log1p( res.x, a.x, xdir );
	mpfr_log1p( res.y, a.y, ydir );
	mpfr_log1p( res.z, a.z, zdir );

	return res;
}

//////////////////////////////////////////////////////////////////////////////

//****f* SAM_math/exp
//    NAME
//       exp
//    SYNOPSIS
//      res =  exp(x)
//    FUNCTION
//      The exp() function computes e**x, the base-e exponential of x.
//
//
//    INPUTS
//      a           - mp_st
//    RESULT
//      res         - mp_st
//    SEE ALSO
//       log() , log2() , log10() , log1p(), exp2(3), expm1(3), pow(3)
//
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//
//
//    BUGS
//
//
//    SEE ALSO
//
//
//  /


mp_st exp(const mp_st& a)
{
	mp_st res;

	if(_SAM_math_tag!=0){
		if (a.accuracy==DIGIT_NOT_COMPUTED) a.nb_significant_digit();
		if (a.accuracy==0) instability(&_SAM_math_count);
	}
	res.accuracy=DIGIT_NOT_COMPUTED;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	mpfr_exp( res.x, a.x, xdir );
	mpfr_exp( res.y, a.y, ydir );
	mpfr_exp( res.z, a.z, zdir );

	return res;
}

//////////////////////////////////////////////////////////////////////////////
//****f* SAM_math/exp2
//    NAME
//       exp2
//    SYNOPSIS
//      res = exp2(x)
//    FUNCTION
//      The exp2() function computes 2**x, the base-2 exponential of
//      x.
//
//
//    INPUTS
//      a           - mp_st
//    RESULT
//      res         - mp_st
//    SEE ALSO
//      log() , log2() , log10() , log1p(), exp(3), expm1(3), pow(3)
//
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//
//
//    BUGS
//
//
//    SEE ALSO
//
//
//  /

mp_st exp2(const mp_st& a)
{
	mp_st res;

	if(_SAM_math_tag!=0){
		if (a.accuracy==DIGIT_NOT_COMPUTED) a.nb_significant_digit();
		if (a.accuracy==0) instability(&_SAM_math_count);
	}
	res.accuracy=DIGIT_NOT_COMPUTED;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	mpfr_exp2( res.x, a.x, xdir );
	mpfr_exp2( res.y, a.y, ydir );
	mpfr_exp2( res.z, a.z, zdir );

	return res;
}

//////////////////////////////////////////////////////////////////////////////

//****f* SAM_math/expm1
//    NAME
//       expm1
//    SYNOPSIS
//      res =  expm1(x)
//    FUNCTION
//      The expm1() function computes the base-e exponential of x ,
//      minus 1 accurately even for very small values of x.
//
//
//    INPUTS
//      a           - mp_st
//    RESULT
//      res         - mp_st
//    SEE ALSO
//      log() , log2() , log10() , log1p(), exp(3), exp2(3), pow(3)
//
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//
//
//    BUGS
//
//
//    SEE ALSO
//
//
//  /

mp_st expm1(const mp_st& a)
{
	mp_st res;

	if(_SAM_math_tag!=0){
		if (a.accuracy==DIGIT_NOT_COMPUTED) a.nb_significant_digit();
		if (a.accuracy==0) instability(&_SAM_math_count);
	}
	res.accuracy=DIGIT_NOT_COMPUTED;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	mpfr_expm1( res.x, a.x, xdir );
	mpfr_expm1( res.y, a.y, ydir );
	mpfr_expm1( res.z, a.z, zdir );

	return res;
}

//////////////////////////////////////////////////////////////////////////////

//****f* SAM_math/sqrt
//    NAME
//      sqrt
//    SYNOPSIS
//      res = sqrt(x)
//    FUNCTION
//      The sqrt() function compute the non-negative square root of x.
//
//
//    INPUTS
//      a           - mp_st
//    RESULT
//      res         - mp_st
//    SEE ALSO
//      cbrt(3)
//
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//
//
//    BUGS
//
//
//
//  /


mp_st sqrt(const mp_st& a)
{
	mp_st res;
	if(_SAM_math_tag!=0){
		if (a.accuracy==DIGIT_NOT_COMPUTED) a.nb_significant_digit();
		if (a.accuracy==0) instability(&_SAM_math_count);
	}
	res.accuracy=DIGIT_NOT_COMPUTED;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	mpfr_sqrt( res.x, a.x, xdir );
	mpfr_sqrt( res.y, a.y, ydir );
	mpfr_sqrt( res.z, a.z, zdir );

	return res;
}

//////////////////////////////////////////////////////////////////////////////


//****f* SAM_math/cbrt
//    NAME
//      cbrt
//    SYNOPSIS
//      res = cbrt(x)
//    FUNCTION
//      The cbrt() function computes the cube root of x.
//
//
//    INPUTS
//      a           - mp_st
//    RESULT
//      res         - mp_st
//    SEE ALSO
//      sqrt(3)
//
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//
//
//    BUGS
//
//
//    SEE ALSO
//
//
//  /

mp_st cbrt(const mp_st& a)
{
	mp_st res;
	if(_SAM_math_tag!=0){
		if (a.accuracy==DIGIT_NOT_COMPUTED) a.nb_significant_digit();
		if (a.accuracy==0) instability(&_SAM_math_count);
	}
	res.accuracy=DIGIT_NOT_COMPUTED;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	mpfr_cbrt( res.x, a.x, xdir );
	mpfr_cbrt( res.y, a.y, ydir );
	mpfr_cbrt( res.z, a.z, zdir );

	return res;
}

//////////////////////////////////////////////////////////////////////////////

//****f* SAM_math/sin
//    NAME
//      sin
//    SYNOPSIS
//      res = sin(x)
//    FUNCTION
//      The sin() function computes the sine of x (measured in
//      radians).
//
//
//    INPUTS
//      a           - mp_st
//    RESULT
//      res         - mp_st
//    SEE ALSO
//      acos(3), asin(3), atan(3), atan2(3), cos(3), cosh(3), sinh(3),
//      tan(3), tanh(3)
//
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//
//
//    BUGS
//
//
//    SEE ALSO
//      acos(3), asin(3), atan(3), atan2(3), cos(3), cosh(3), sinh(3),
//      tan(3), tanh(3)
//
//  /

mp_st sin(const mp_st& a)
{
	mp_st res;
	res.accuracy=DIGIT_NOT_COMPUTED;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	mpfr_sin( res.x, a.x, xdir );
	mpfr_sin( res.y, a.y, ydir );
	mpfr_sin( res.z, a.z, zdir );

	return res;
}

//////////////////////////////////////////////////////////////////////////////

//****f* SAM_math/cos
//    NAME
//      cos
//    SYNOPSIS
//      res = cos(x)
//    FUNCTION
//      The cos() function computes the cosine of x (measured in
//      radians).
//
//
//    INPUTS
//      a           - mp_st
//    RESULT
//      res         - mp_st
//    SEE ALSO
//      acos(3), asin(3), atan(3), atan2(3), sin(3), cosh(3), sinh(3),
//      tan(3), tanh(3)
//
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//
//
//    BUGS
//
//
//    SEE ALSO
//
//
//  /

mp_st cos(const mp_st& a)
{
	mp_st res;
	res.accuracy=DIGIT_NOT_COMPUTED;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	mpfr_cos( res.x, a.x, xdir );
	mpfr_cos( res.y, a.y, ydir );
	mpfr_cos( res.z, a.z, zdir );

	return res;
}

//////////////////////////////////////////////////////////////////////////////

//****f* SAM_math/tan
//    NAME
//      tan
//    SYNOPSIS
//      res = tan(x)
//    FUNCTION
//      The tan() function computes the tangent of x (measured in
//      radians).
//
//
//    INPUTS
//      a           - mp_st
//    RESULT
//      res         - mp_st
//    SEE ALSO
//      acos(3), asin(3), atan(3), atan2(3), cos(3), cosh(3), sinh(3),
//      sin(3), tanh(3)
//
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//
//
//    BUGS
//
//
//    SEE ALSO
//
//
//  /


mp_st tan(const mp_st& a)
{
	mp_st res;
	res.accuracy=DIGIT_NOT_COMPUTED;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	mpfr_tan( res.x, a.x, xdir );
	mpfr_tan( res.y, a.y, ydir );
	mpfr_tan( res.z, a.z, zdir );

	return res;
}

//////////////////////////////////////////////////////////////////////////////

//****f* SAM_math/asin
//    NAME
//      asin
//    SYNOPSIS
//      res = asin(x)
//    FUNCTION
//      The asin() function computes the principal value of the arc
//      sine of x.  The result is in the range [-pi/2, +pi/2].
//
//
//    INPUTS
//      a           - mp_st
//    RESULT
//      res         - mp_st
//    SEE ALSO
//      acos(3), sin(3), atan(3), atan2(3), cos(3), cosh(3), sinh(3),
//      tan(3), tanh(3)
//
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//
//
//    BUGS
//
//
//    SEE ALSO
//
//
//  /

mp_st asin(const mp_st& a)
{
	mp_st res;
	res.accuracy=DIGIT_NOT_COMPUTED;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	mpfr_asin( res.x, a.x, xdir );
	mpfr_asin( res.y, a.y, ydir );
	mpfr_asin( res.z, a.z, zdir );

	return res;
}

//////////////////////////////////////////////////////////////////////////////

//****f* SAM_math/acos
//    NAME
//      acos
//    SYNOPSIS
//      res = acos(x)
//    FUNCTION
//      The acos() function computes the principle value of the arc
//      cosine of x.  The result is in the range [0, pi].
//
//
//    INPUTS
//      a           - mp_st
//    RESULT
//      res         - mp_st
//    SEE ALSO
//      sin(3), asin(3), atan(3), atan2(3), cos(3), cosh(3), sinh(3),
//      tan(3), tanh(3)
//
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//
//
//    BUGS
//
//
//    SEE ALSO
//
//
//  /

mp_st acos(const mp_st& a)
{
	mp_st res;
	res.accuracy=DIGIT_NOT_COMPUTED;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	mpfr_acos( res.x, a.x, xdir );
	mpfr_acos( res.y, a.y, ydir );
	mpfr_acos( res.z, a.z, zdir );

	return res;
}

//////////////////////////////////////////////////////////////////////////////

//****f* SAM_math/atan
//    NAME
//      atan
//    SYNOPSIS
//      res = atan(x)
//    FUNCTION
//      The atan() function computes the principal value of the arc
//      tangent of x.  The result is in the range [-pi/2, +pi/2].
//
//
//    INPUTS
//      a           - mp_st
//    RESULT
//      res         - mp_st
//    SEE ALSO
//      acos(3), asin(3), sin(3), atan2(3), cos(3), cosh(3), sinh(3),
//      tan(3), tanh(3)
//
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//
//
//    BUGS
//
//
//    SEE ALSO
//
//
//  /

mp_st atan(const mp_st& a)
{
	mp_st res;
	res.accuracy=DIGIT_NOT_COMPUTED;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	mpfr_atan( res.x, a.x, xdir );
	mpfr_atan( res.y, a.y, ydir );
	mpfr_atan( res.z, a.z, zdir );

	return res;
}

//////////////////////////////////////////////////////////////////////////////



//****f* SAM_math/atan2
//    NAME
//      atan2
//    SYNOPSIS
//      res = atan2(x,y)
//    FUNCTION
//      The atan2() function computes the principal value of the arc
//      tangent of y/x, using the signs of both arguments to determine
//      the quadrant of the return value.
//
//
//    INPUTS
//      x           - mp_st
//      y           - mp_st
//    RESULT
//      res         - mp_st
//    SEE ALSO
//      acos(3), asin(3), atan(3), sin(3), cos(3), cosh(3), sinh(3),
//      tan(3), tanh(3)
//
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//
//
//    BUGS
//
//
//    SEE ALSO
//
//
//  /

mp_st atan2(const mp_st& a, const mp_st& b)
{
	mp_st res;
	res.accuracy=DIGIT_NOT_COMPUTED;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	mpfr_atan2( res.x, a.x, b.x, xdir );
	mpfr_atan2( res.y, a.y, b.y, ydir );
	mpfr_atan2( res.z, a.z, b.z, zdir );

	return res;
}

//////////////////////////////////////////////////////////////////////////////

//****f* SAM_math/sinh
//    NAME
//      sinh
//    SYNOPSIS
//      res = sinh(x)
//    FUNCTION
//      The sinh() function computes the hyperbolic sine of x.
//
//
//    INPUTS
//      a           - mp_st
//    RESULT
//      res         - mp_st
//    SEE ALSO
//      acos(3), asin(3), atan(3), atan2(3), cos(3), cosh(3), sin(3),
//      tan(3), tanh(3)
//
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//
//
//    BUGS
//
//
//    SEE ALSO
//
//
//  /

mp_st sinh(const mp_st& a)
{
	mp_st res;
	res.accuracy=DIGIT_NOT_COMPUTED;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	mpfr_sinh( res.x, a.x, xdir );
	mpfr_sinh( res.y, a.y, ydir );
	mpfr_sinh( res.z, a.z, zdir );

	return res;
}

//////////////////////////////////////////////////////////////////////////////

//****f* SAM_math/cosh
//    NAME
//      cosh
//    SYNOPSIS
//      res = cosh(x)
//    FUNCTION
//      The cosh() function computes the hyperbolic cosine of x.
//
//
//    INPUTS
//      a           - mp_st
//    RESULT
//      res         - mp_st
//    SEE ALSO
//      acos(3), asin(3), atan(3), atan2(3), cos(3), sin(3), sinh(3),
//      tan(3), tanh(3)
//
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//
//
//    BUGS
//
//
//    SEE ALSO
//
//
//  /

mp_st cosh(const mp_st& a)
{
	mp_st res;
	res.accuracy=DIGIT_NOT_COMPUTED;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	mpfr_cosh( res.x, a.x, xdir );
	mpfr_cosh( res.y, a.y, ydir );
	mpfr_cosh( res.z, a.z, zdir );

	return res;
}

//////////////////////////////////////////////////////////////////////////////

//****f* SAM_math/tanh
//    NAME
//      tanh
//    SYNOPSIS
//      res = tanh(x)
//    FUNCTION
//      The tanh() function computes the hyperbolic tangent of x.
//
//
//    INPUTS
//      a           - mp_st
//    RESULT
//      res         - mp_st
//    SEE ALSO
//      acos(3), asin(3), atan(3), atan2(3), cos(3), cosh(3), sinh(3),
//      tan(3), sin(3)
//
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//
//
//    BUGS
//
//
//    SEE ALSO
//
//
//  /

mp_st tanh(const mp_st& a)
{
	mp_st res;
	res.accuracy=DIGIT_NOT_COMPUTED;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	mpfr_tanh( res.x, a.x, xdir );
	mpfr_tanh( res.y, a.y, ydir );
	mpfr_tanh( res.z, a.z, zdir );

	return res;
}

//////////////////////////////////////////////////////////////////////////////

//****f* SAM_math/asinh
//    NAME
//      asinh
//    SYNOPSIS
//      res = asinh(x)
//    FUNCTION
//      The asinh() function computes the inverse hyperbolic sine of x.
//
//
//
//    INPUTS
//      a           - mp_st
//    RESULT
//      res         - mp_st
//    SEE ALSO
//      acos(3), asin(3), atan(3), atan2(3), cos(3), cosh(3), sinh(3),
//      tan(3), tanh(3)
//
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//
//
//    BUGS
//
//
//    SEE ALSO
//
//
//  /

mp_st asinh(const mp_st& a)
{
	mp_st res;
	res.accuracy=DIGIT_NOT_COMPUTED;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	mpfr_asinh( res.x, a.x, xdir );
	mpfr_asinh( res.y, a.y, ydir );
	mpfr_asinh( res.z, a.z, zdir );

	return res;
}

//////////////////////////////////////////////////////////////////////////////

//****f* SAM_math/acosh
//    NAME
//      acosh
//    SYNOPSIS
//      res = acosh(x)
//    FUNCTION
//      The acosh() function computes the inverse hyperbolic cosine of x.
//
//
//
//    INPUTS
//      a           - mp_st
//    RESULT
//      res         - mp_st
//    SEE ALSO
//      acos(3), asin(3), atan(3), atan2(3), cos(3), cosh(3), sinh(3),
//      tan(3), tanh(3)
//
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//
//
//    BUGS
//
//
//    SEE ALSO
//
//
//  /

mp_st acosh(const mp_st& a)
{
	mp_st res;
	res.accuracy=DIGIT_NOT_COMPUTED;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	mpfr_acosh( res.x, a.x, xdir );
	mpfr_acosh( res.y, a.y, ydir );
	mpfr_acosh( res.z, a.z, zdir );

	return res;
}

//////////////////////////////////////////////////////////////////////////////

//****f* SAM_math/atanh
//    NAME
//      atanh
//    SYNOPSIS
//      res = atanh(x)
//    FUNCTION
//      The atanh() function computes the inverse hyperbolic tangent
//      of x.
//
//
//    INPUTS
//      a           - mp_st
//    RESULT
//      res         - mp_st
//    SEE ALSO
//      acos(3), asin(3), atan(3), atan2(3), cos(3), cosh(3), sinh(3),
//      tan(3), tanh(3)
//
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//
//
//    BUGS
//
//
//    SEE ALSO
//
//
//  /

mp_st atanh(const mp_st& a)
{
	mp_st res;
	res.accuracy=DIGIT_NOT_COMPUTED;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	mpfr_atanh( res.x, a.x, xdir );
	mpfr_atanh( res.y, a.y, ydir );
	mpfr_atanh( res.z, a.z, zdir );

	return res;
}

//////////////////////////////////////////////////////////////////////////////
//****f* SAM_math/hypot
//    NAME
//      hypot
//    SYNOPSIS
//      res = hypot(x,y)
//    FUNCTION
//      The hypot() function computes the sqrt(x*x+y*y) without undue
//      overflow or underflow.
//
//
//    INPUTS
//      a           - mp_st
//      b           - mp_st
//    RESULT
//      res         - mp_st
//    SEE ALSO
//      sqrtf(3)
//
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//
//
//    BUGS
//
//
//    SEE ALSO
//
// /

mp_st hypot(const mp_st& a, const mp_st& b)
{
	mp_st res;
	res.accuracy=DIGIT_NOT_COMPUTED;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	mpfr_hypot( res.x, a.x, b.x, xdir );
	mpfr_hypot( res.y, a.y, b.y, ydir );
	mpfr_hypot( res.z, a.z, b.z, zdir );

	return res;
}







