/*-----------------------------------------------------------------------
Copyright 2010-2016 S. GRAILLAT, F. JEZEQUEL
Copyright 2010 S. WANG, Y. ZHU
Copyright 2019-2020 J.E. TIDSWELL

This file is part of SAM.

    SAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    SAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
    Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with SAM. If not, see
    <http://www.gnu.org/licenses/>.
-----------------------------------------------------------------------*/
///////////////////////////////////////////////////
//
// SAM SOFTWARE
//
///////////////////////////////////////////////////
// Last Modifications
//
//
///////////////////////////////////////////////////
//****m* SAM_mul/operator*
//    NAME
//      operator*
//    SYNOPSIS
//      res = a * b
//    FUNCTION
//      Defines all the functions involving at least one argument
//      of stochastic type which overload the "*" operator
//      in a statement such as "a*b".
//
//    INPUTS
//      a           - an integer, a float, a double, an mpfr_t object or a stochastic number
//      b           - an integer, a float, a double, an mpfr_t object or a stochastic number
//      At least one argument must be of stochastic type.
//    RESULT
//      res         - a stochastic number
//
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//
//
//    BUGS
//
//
//    SEE ALSO
//
//
//  /


#include "SAM.h"
#include "SAM_private.h"
#include "SAM_MPFR_templates.h"

using namespace SAM;


///////////////////////////////////////////////////////////////////////////////////////////
// operator *=()

mp_st& mp_st::operator*=(const mp_st& a)
{
	if(_SAM_mul_tag) {
		if (accuracy==DIGIT_NOT_COMPUTED) nb_significant_digit();
		if (accuracy==0){
			if (a.accuracy==DIGIT_NOT_COMPUTED) a.nb_significant_digit();
			if (a.accuracy==0) {
				instability(&_SAM_mul_count);
			}
		}
	}

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	SAM::mul_mpfr( x, a.x, xdir );
	SAM::mul_mpfr( y, a.y, ydir );
	SAM::mul_mpfr( z, a.z, zdir );

	if (_SAM_cancel_tag) {
		this->accuracy = this->nb_significant_digit(); // preemptive calculate
	} else {
		this->accuracy = DIGIT_NOT_COMPUTED;
	}
	return *this;
}


template<typename argT> mp_st& mp_st::operator*=(const argT& a)
{
	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	SAM::mul_mpfr( x, a, xdir );
	SAM::mul_mpfr( y, a, ydir );
	SAM::mul_mpfr( z, a, zdir );

	if (_SAM_cancel_tag) {
		this->accuracy = this->nb_significant_digit(); // preemptive calculate
	} else {
		this->accuracy = DIGIT_NOT_COMPUTED;
	}
	return *this;
}


//
// explicit instantiation so compiler generates code to include in linker
//
template mp_st& mp_st::operator*=<int>(const int&);
template mp_st& mp_st::operator*=<unsigned int>(const unsigned int&);
template mp_st& mp_st::operator*=<long>(const long&);
template mp_st& mp_st::operator*=<unsigned long>(const unsigned long&);
template mp_st& mp_st::operator*=<float>(const float&);
template mp_st& mp_st::operator*=<double>(const double&);
template mp_st& mp_st::operator*=<mpfr_t>(const mpfr_t&);
template mp_st& mp_st::operator*=<mpfr_srcptr>(const mpfr_srcptr&);


