/*-----------------------------------------------------------------------
Copyright 2019-2020 J.E. TIDSWELL

This file is part of SAM.

    SAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    SAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
    Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with SAM. If not, see
    <http://www.gnu.org/licenses/>.
-----------------------------------------------------------------------*/
///////////////////////////////////////////////////
//
// SAM SOFTWARE
// 
///////////////////////////////////////////////////
// Last Modifications
// 
//
///////////////////////////////////////////////////
//****m* SAM_ordering
//    NAME
//      order(a,b)
//    SYNOPSIS
//      r = -1 if a < b, +1 if a > b, 0 otherwise 
//    FUNCTION
//      Define ordering involvoing mp_st values as basis for relational operators.
//      
//    INPUTS
//      a           - an integer, a float, a double, an mpfr_t object or a stochastic number
//      b           - an integer, a float, a double, an mpfr_t object or a stochastic number
//      At least one argument must be of stochastic type.
//    RESULT
//      r           - an integer value
//             
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//      
//  
//    NOTES
//    The output value takes into account the stochastic definition of
//    the relational operators (see the documentation).
//  
//    BUGS
//  
//  
//    SEE ALSO
//      
//      
//  /


#include <limits>

#include "SAM.h"
#include "SAM_private.h"
#include "SAM_MPFR_templates.h"

using namespace SAM;

///////////////////////////////////////////////////
//****m* SAM_ordering/order -
//    NAME
//      order -
//    SYNOPSIS
//      res := order( a, b )
//    FUNCTION
//      Defines an ordering relation on pair of arguments (as long as one is a stochastic number).
//    INPUTS
//      a           - an integer, a float, a double, an mpfr_t object or a stochastic number
//      b           - an integer, a float, a double, an mpfr_t object or a stochastic number
//      At least one argument must be of stochastic type.
//    RESULT
//      res         -  an integer in set { -1, 0, +1 } where 
//            -1 implies a < b
//            +1 implies a > b
//             0 implies either a == b (exact equality)
//                           or a-b (difference) is stochastic noise (which is counted as a branching instabilibity)
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//
//
//    NOTES
//         Ordering is formally defined using both deterministic and stochastic differences.
//         This is costly to calculate, so attempt to avoid calculating both unless necessary.
//
//         The basic pattern is as follows:
//          1. calculate difference using stochastic arithmetic;
//          2. if stochastic difference is an exact zero the result is zero;
//          3. if stochastic difference is "noise" then result is still zero, but report instability (lack certainty);
//          4. if the stochastic difference is large, then the randomisation will not have affected sign,
//             so the result is the sign of the stochastic difference;
//          5. if the stochastic difference is small, then perhaps the randomisation affected the sign of the answer,
//             so calculate the deterministic difference and use the sign of this.
//
//    BUGS
//
//
//    SEE ALSO
//
//
//  /

///////////////////////////////////////////////////////////////////////////////////////////

int order(const mp_st& a, const mp_st& b)
{
	mp_st tmp;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	SAM::sub_mpfr( tmp.x , a.x, b.x, xdir );
	SAM::sub_mpfr( tmp.y , a.y, b.y, ydir );
	SAM::sub_mpfr( tmp.z , a.z, b.z, zdir );

	int r = 0; // computed zero is default
	
	if ( (mpfr_get_kind(tmp.x) == MPFR_ZERO_KIND)  && (mpfr_get_kind(tmp.x) == MPFR_ZERO_KIND) && (mpfr_get_kind(tmp.x) == MPFR_ZERO_KIND) ) {
		/* do nothing */
	} 
	else if ( tmp.computedzero() ) {
		/* not exact, so its noise */
		if ( SAM::_SAM_branching_tag ) { SAM::instability(&SAM::_SAM_branching_count); }
	}
	else {
		// test if randomised error MIGHT affect sign
		long kdi, kab;
		const mpfr_prec_t p = DEFAULT_PREC;
		bool calcMeanDiff = false;
		kdi = mpfr_get_exp(tmp.x);
		kab = std::max(mpfr_get_exp(a.x), mpfr_get_exp(b.x));
		calcMeanDiff |= (kdi>kab-p);
		kdi = mpfr_get_exp(tmp.y);
		kab = std::max(mpfr_get_exp(a.y), mpfr_get_exp(b.y));
		calcMeanDiff |= (kdi>kab-p);
		kdi = mpfr_get_exp(tmp.z);
		kab = std::max(mpfr_get_exp(a.z), mpfr_get_exp(b.z));
		calcMeanDiff |= (kdi>kab-p);
		if ( !calcMeanDiff ) {
			// randomised errors in stochastic difference small enough
			// so use stochastic difference to determine order
			r = mpfr_sgn(tmp.x);
		}
		else {
			// randomised errors in stochastic difference too large
			// calculate order long way to ovoid overflow
			// reuse tmp to avoid additional heap allocation for totals
			SAM::sub_mpfr (tmp.x, a.x, b.x, MPFR_RNDN);
			SAM::sub_mpfr (tmp.y, a.y, b.y, MPFR_RNDN);
			SAM::sub_mpfr (tmp.z, a.z, b.z, MPFR_RNDN);
			SAM::add_mpfr (tmp.x, tmp.x, tmp.y, MPFR_RNDN);
			SAM::add_mpfr (tmp.x, tmp.x, tmp.z, MPFR_RNDN);
			r = mpfr_sgn(tmp.x);
		}
	}
	return r;
}


///////////////////////////////////////////////////////////////////////////////////////////

int order(const mpfr_t& a, const mp_st& b)
{
	mp_st tmp;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	SAM::sub_mpfr( tmp.x , a, b.x, xdir );
	SAM::sub_mpfr( tmp.y , a, b.y, ydir );
	SAM::sub_mpfr( tmp.z , a, b.z, zdir );

	int r = 0; // computed zero is default
	
	if ( (mpfr_get_kind(tmp.x) == MPFR_ZERO_KIND)  && (mpfr_get_kind(tmp.x) == MPFR_ZERO_KIND) && (mpfr_get_kind(tmp.x) == MPFR_ZERO_KIND) ) {
		/* do nothing */
	} 
	else if ( tmp.computedzero() ) {
		/* not exact, so its noise */
		if ( SAM::_SAM_branching_tag ) { SAM::instability(&SAM::_SAM_branching_count); }
	}
	else {
		// test if randomised error MIGHT affect sign
		long kdi, ka, kab;
		const mpfr_prec_t p = DEFAULT_PREC;
		bool calcMeanDiff = false;
		kdi = mpfr_get_exp(tmp.x);
		ka = mpfr_get_exp(a);
		kab = std::max(ka, mpfr_get_exp(b.x));
		calcMeanDiff |= (kdi>kab-p);
		kdi = mpfr_get_exp(tmp.y);
		kab = std::max(ka, mpfr_get_exp(b.y));
		calcMeanDiff |= (kdi>kab-p);
		kdi = mpfr_get_exp(tmp.z);
		kab = std::max(ka, mpfr_get_exp(b.z));
		calcMeanDiff |= (kdi>kab-p);
		if ( !calcMeanDiff ) {
			// randomised errors in stochastic difference small enough
			// so use stochastic difference to determine order
			r = mpfr_sgn(tmp.x);
		}
		else {
			// randomised errors in stochastic difference too large
			// calculate order long way to ovoid overflow
			// reuse tmp to avoid additional heap allocation for totals
			SAM::sub_mpfr (tmp.x, a, b.x, MPFR_RNDN);
			SAM::sub_mpfr (tmp.y, a, b.y, MPFR_RNDN);
			SAM::sub_mpfr (tmp.z, a, b.z, MPFR_RNDN);
			SAM::add_mpfr (tmp.x, tmp.x, tmp.y, MPFR_RNDN);
			SAM::add_mpfr (tmp.x, tmp.x, tmp.z, MPFR_RNDN);
			r = mpfr_sgn(tmp.x);
		}
	}

	return r;
}


///////////////////////////////////////////////////////////////////////////////////////////

int order(const mp_st& a, const mpfr_t& b)
{
	mp_st tmp;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	SAM::sub_mpfr( tmp.x , a.x, b, xdir );
	SAM::sub_mpfr( tmp.y , a.y, b, ydir );
	SAM::sub_mpfr( tmp.z , a.z, b, zdir );

	int r = 0; // computed zero is default
	
	if ( (mpfr_get_kind(tmp.x) == MPFR_ZERO_KIND)  && (mpfr_get_kind(tmp.x) == MPFR_ZERO_KIND) && (mpfr_get_kind(tmp.x) == MPFR_ZERO_KIND) ) {
		/* do nothing */
	} 
	else if ( tmp.computedzero() ) {
		/* not exact, so its noise */
		if ( SAM::_SAM_branching_tag ) { SAM::instability(&SAM::_SAM_branching_count); }
	}
	else {
		// test if randomised error MIGHT affect sign
		long kdi, kb, kab;
		const mpfr_prec_t p = DEFAULT_PREC;
		bool calcMeanDiff = false;
		kdi = mpfr_get_exp(tmp.x);
		kb = mpfr_get_exp(b);
		kab = std::max(mpfr_get_exp(a.x),kb);
		calcMeanDiff |= (kdi>kab-p);
		kdi = mpfr_get_exp(tmp.y);
		kab = std::max(mpfr_get_exp(a.y),kb);
		calcMeanDiff |= (kdi>kab-p);
		kdi = mpfr_get_exp(tmp.z);
		kab = std::max(mpfr_get_exp(a.z),kb);
		calcMeanDiff |= (kdi>kab-p);
		if ( !calcMeanDiff ) {
			// randomised errors in stochastic difference small enough
			// so use stochastic difference to determine order
			r = mpfr_sgn(tmp.x);
		}
		else {
			// randomised errors in stochastic difference too large
			// calculate order long way to ovoid overflow
			// reuse tmp to avoid additional heap allocation for totals
			SAM::sub_mpfr (tmp.x, a.x, b, MPFR_RNDN);
			SAM::sub_mpfr (tmp.y, a.y, b, MPFR_RNDN);
			SAM::sub_mpfr (tmp.z, a.z, b, MPFR_RNDN);
			SAM::add_mpfr (tmp.x, tmp.x, tmp.y, MPFR_RNDN);
			SAM::add_mpfr (tmp.x, tmp.x, tmp.z, MPFR_RNDN);
			r = mpfr_sgn(tmp.x);
		}
	}

	return r; 
}


///////////////////////////////////////////////////////////////////////////////////////////

// utility function for exponents

//static inline long get_exponent(mpfr_srcptr b) { return mpfr_get_exp(b); }
static inline long get_exponent(double b) { return ilogb(b); }
static inline long get_exponent(float b) { return ilogbf(b); }
static inline long get_exponent(unsigned long b) { return ilogb((double)b); }
static inline long get_exponent(long b) { return ilogb((double)b); }
static inline long get_exponent(unsigned int b) { return ilogb((double)b); }
static inline long get_exponent(int b) { return ilogb((double)b); }

///////////////////////////////////////////////////////////////////////////////////////////

template<typename argT> 
int order(const mp_st& a, const argT& b)
{
	mp_st tmp;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	SAM::sub_mpfr( tmp.x , a.x, b, xdir );
	SAM::sub_mpfr( tmp.y , a.y, b, ydir );
	SAM::sub_mpfr( tmp.z , a.z, b, zdir );

	int r = 0; // computed zero is default
	
	if ( (mpfr_get_kind(tmp.x) == MPFR_ZERO_KIND)  && (mpfr_get_kind(tmp.x) == MPFR_ZERO_KIND) && (mpfr_get_kind(tmp.x) == MPFR_ZERO_KIND) ) {
		/* do nothing */
	} 
	else if ( tmp.computedzero() ) {
		/* not exact, so its noise */
		if ( SAM::_SAM_branching_tag ) { SAM::instability(&SAM::_SAM_branching_count); }
	}
	else {
		// test if randomised error MIGHT affect sign
		long kdi, kb, kab;
		const mpfr_prec_t p = DEFAULT_PREC;
		bool calcMeanDiff = false;
		kdi = mpfr_get_exp(tmp.x);
		kb = get_exponent(b);
		kab = std::max(mpfr_get_exp(a.x),kb);
		calcMeanDiff |= (kdi>kab-p);
		kdi = mpfr_get_exp(tmp.y);
		kab = std::max(mpfr_get_exp(a.y),kb);
		calcMeanDiff |= (kdi>kab-p);
		kdi = mpfr_get_exp(tmp.z);
		kab = std::max(mpfr_get_exp(a.z),kb);
		calcMeanDiff |= (kdi>kab-p);
		if ( !calcMeanDiff ) {
			// randomised errors in stochastic difference small enough
			// so use stochastic difference to determine order
			r = mpfr_sgn(tmp.x);
		}
		else {
			// randomised errors in stochastic difference too large
			// calculate order long way to ovoid overflow
			// reuse tmp to avoid additional heap allocation for totals
			SAM::sub_mpfr (tmp.x, a.x, b, MPFR_RNDN);
			SAM::sub_mpfr (tmp.y, a.y, b, MPFR_RNDN);
			SAM::sub_mpfr (tmp.z, a.z, b, MPFR_RNDN);
			SAM::add_mpfr (tmp.x, tmp.x, tmp.y, MPFR_RNDN);
			SAM::add_mpfr (tmp.x, tmp.x, tmp.z, MPFR_RNDN);
			r = mpfr_sgn(tmp.x);
		}
	}
	return r;
}


template<typename argT> 
int order(const argT& a, const mp_st& b)
{
	mp_st tmp;

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	SAM::sub_mpfr( tmp.x, a, b.x, xdir );
	SAM::sub_mpfr( tmp.y, a, b.y, ydir );
	SAM::sub_mpfr( tmp.z, a, b.z, zdir );

	int r = 0; // computed zero is default
	
	if ( (mpfr_get_kind(tmp.x) == MPFR_ZERO_KIND)  && (mpfr_get_kind(tmp.x) == MPFR_ZERO_KIND) && (mpfr_get_kind(tmp.x) == MPFR_ZERO_KIND) ) {
		/* do nothing */
	} 
	else if ( tmp.computedzero() ) {
		/* not exact, so its noise */
		if ( SAM::_SAM_branching_tag ) { SAM::instability(&SAM::_SAM_branching_count); }
	}
	else {
		// test if randomised error MIGHT affect sign
		long kdi, ka, kab;
		const mpfr_prec_t p = DEFAULT_PREC;
		bool calcMeanDiff = false;
		kdi = mpfr_get_exp(tmp.x);
		ka = get_exponent(a);
		kab = std::max(ka, mpfr_get_exp(b.x));
		calcMeanDiff |= (kdi>kab-p);
		kdi = mpfr_get_exp(tmp.y);
		kab = std::max(ka, mpfr_get_exp(b.y));
		calcMeanDiff |= (kdi>kab-p);
		kdi = mpfr_get_exp(tmp.z);
		kab = std::max(ka, mpfr_get_exp(b.z));
		calcMeanDiff |= (kdi>kab-p);
		if ( !calcMeanDiff ) {
			// randomised errors in stochastic difference small enough
			// so use stochastic difference to determine order
			r = mpfr_sgn(tmp.x);
		}
		else {
			// randomised errors in stochastic difference too large
			// calculate order long way to ovoid overflow
			// reuse tmp to avoid additional heap allocation for totals
			SAM::sub_mpfr (tmp.x, a, b.x, MPFR_RNDN);
			SAM::sub_mpfr (tmp.y, a, b.y, MPFR_RNDN);
			SAM::sub_mpfr (tmp.z, a, b.z, MPFR_RNDN);
			SAM::add_mpfr (tmp.x, tmp.x, tmp.y, MPFR_RNDN);
			SAM::add_mpfr (tmp.x, tmp.x, tmp.z, MPFR_RNDN);
			r = mpfr_sgn(tmp.x);
		}
	}
	return r;
}


//
// explicit instantiation so compiler generates code to include in linker
//

template int order<>(const mp_st&, const int&);
template int order<>(const mp_st&, const unsigned int&);
template int order<>(const mp_st&, const long&);
template int order<>(const mp_st&, const unsigned long&);
template int order<>(const mp_st&, const float&);
template int order<>(const mp_st&, const double&);
// template int order<>(const mp_st&, const mpfr_t&);
// template int order<>(const mp_st&, const mpfr_srcptr&);

template int order<>(const int&, const mp_st&);
template int order<>(const unsigned int&, const mp_st&);
template int order<>(const long&, const mp_st&);
template int order<>(const unsigned long&, const mp_st&);
template int order<>(const float&, const mp_st&);
template int order<>(const double&, const mp_st&);
// template int order<>(const mpfr_t&, const mp_st&);
// template int order<>(const mpfr_srcptr&, const mp_st&);



///////////////////////////////////////////////////////////////////////////////////////////

int identical(const mp_st& a, const mp_st& b)
{
	int xmatch = (SAM::cmp_mpfr(a.x, b.x)==0) || ( mpfr_nan_p(a.x) && mpfr_nan_p(b.x) );
	int ymatch = (SAM::cmp_mpfr(a.y, b.y)==0) || ( mpfr_nan_p(a.y) && mpfr_nan_p(b.y) );
	int zmatch = (SAM::cmp_mpfr(a.z, b.z)==0) || ( mpfr_nan_p(a.z) && mpfr_nan_p(b.z) );
	return xmatch && ymatch && zmatch;
}


///////////////////////////////////////////////////////////////////////////////////////////

int identical(const mp_st& a, const mpfr_t& b)
{
	if ( mpfr_nan_p(b) )
	{
		return mpfr_nan_p(a.x) && mpfr_nan_p(a.y) && mpfr_nan_p(a.z);
	}
	else
	{
		int xmatch = (SAM::cmp_mpfr(a.x, b)==0);
		int ymatch = (SAM::cmp_mpfr(a.y, b)==0);
		int zmatch = (SAM::cmp_mpfr(a.z, b)==0);
		return xmatch && ymatch && zmatch;
	}
}


///////////////////////////////////////////////////////////////////////////////////////////

template<typename argT> 
int identical(const mp_st& a, const argT& b)
{
	assert(!std::numeric_limits<argT>::has_quiet_NaN);
	assert(!std::numeric_limits<argT>::has_signaling_NaN);
	int xmatch = (SAM::cmp_mpfr(a.x, b)==0);
	int ymatch = (SAM::cmp_mpfr(a.y, b)==0);
	int zmatch = (SAM::cmp_mpfr(a.z, b)==0);
	return xmatch && ymatch && zmatch;
}


///////////////////////////////////////////////////////////////////////////////////////////

template<>
int identical<>(const mp_st& a, const double& b)
{
	if ( std::isnan(b) )
	{
		return mpfr_nan_p(a.x) && mpfr_nan_p(a.y) && mpfr_nan_p(a.z);
	}
	else
	{
		int xmatch = (SAM::cmp_mpfr(a.x, b)==0);
		int ymatch = (SAM::cmp_mpfr(a.y, b)==0);
		int zmatch = (SAM::cmp_mpfr(a.z, b)==0);
		return xmatch && ymatch && zmatch;
	}
}


///////////////////////////////////////////////////////////////////////////////////////////

template<>
int identical<>(const mp_st& a, const float& b)
{
	if ( std::isnan(b) )
	{
		return mpfr_nan_p(a.x) && mpfr_nan_p(a.y) && mpfr_nan_p(a.z);
	}
	else
	{
		int xmatch = (SAM::cmp_mpfr(a.x, b)==0);
		int ymatch = (SAM::cmp_mpfr(a.y, b)==0);
		int zmatch = (SAM::cmp_mpfr(a.z, b)==0);
		return xmatch && ymatch && zmatch;
	}
}





//
// explicit instantiation so compiler generates code to include in linker
//

template int identical<>(const mp_st&, const int&);
template int identical<>(const mp_st&, const unsigned int&);
template int identical<>(const mp_st&, const long&);
template int identical<>(const mp_st&, const unsigned long&);
// template int identical<>(const mp_st&, const float&);
// template int identical<>(const mp_st&, const double&);
// template int identical<>(const mp_st&, const mpfr_t&);
// template int identical<>(const mp_st&, const mpfr_srcptr&);






