/*-----------------------------------------------------------------------
Copyright 2013 S. GRAILLAT, F. JEZEQUEL, S. WANG, Y. ZHU
Copyright 2020 J. Tidswell

This file is part of SAM.

	SAM is free software: you can redistribute it and/or modify it
	under the terms of the GNU Lesser General Public License as
	published by the Free Software Foundation, either version 3 of the
	License, or (at your option) any later version.

	SAM is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
	or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
	Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with SAM. If not, see
	<http://www.gnu.org/licenses/>.
-----------------------------------------------------------------------*/
#ifndef __SAM_RELATIONS__
#define __SAM_RELATIONS__

#include "SAM_private.h"

//
// relational operators defined using order()
//


// operator ==

inline int operator==(const mp_st& a, const mp_st& b)
{
	return (order(a,b)==0);
}

template<typename argT>
inline int operator==(const mp_st& a, const argT& b)
{
	return (order(a,b)==0);
}

template<typename argT>
inline int operator==(const argT& a, const mp_st& b)
{
	return (order(a,b)==0);
}

// operator !=

inline int operator!=(const mp_st& a, const mp_st& b)
{
	return order(a,b)!=0;
}

template<typename argT>
inline int operator!=(const mp_st& a, const argT& b)
{
	return order(a,b)!=0;
}

template<typename argT>
inline int operator!=(const argT& a, const mp_st& b)
{
	return order(a,b)!=0;
}

// operator >=

inline int operator>=(const mp_st& a, const mp_st& b)
{
	return (order(a,b)>=0);
}

inline int operator>=(const mp_st& a, const mpfr_t& b)
{
	return (order(a,b)>=0);
}

inline int operator>=(const mpfr_t& a, const mp_st& b)
{
	return (order(a,b)>=0);
}

template<typename argT>
inline int operator>=(const mp_st& a, const argT& b)
{
	return (order(a,b)>=0);
}

template<typename argT>
inline int operator>=(const argT& a, const mp_st& b)
{
	return (order(a,b)>=0);
}

// operator >

inline int operator>(const mp_st& a, const mp_st& b)
{
	return (order(a,b)>0);
}

inline int operator>(const mp_st& a, const mpfr_t& b)
{
	return (order(a,b)>0);
}

inline int operator>(const mpfr_t& a, const mp_st& b)
{
	return (order(a,b)>0);
}

template<typename argT>
inline int operator>(const mp_st& a, const argT& b)
{
	return (order(a,b)>0);
}

template<typename argT>
inline int operator>(const argT& a, const mp_st& b)
{
	return (order(a,b)>0);
}

// operator <=

inline int operator<=(const mp_st& a, const mp_st& b)
{
	return (order(a,b)<=0);
}

inline int operator<=(const mp_st& a, const mpfr_t& b)
{
	return (order(a,b)<=0);
}

inline int operator<=(const mpfr_t& a, const mp_st& b)
{
	return (order(a,b)<=0);
}

template<typename argT>
inline int operator<=(const mp_st& a, const argT& b)
{
	return (order(a,b)<=0);
}

template<typename argT>
inline int operator<=(const argT& a, const mp_st& b)
{
	return (order(a,b)<=0);
}

// operator <

inline int operator<(const mp_st& a, const mp_st& b)
{
	return (order(a,b)<0);
}

inline int operator<(const mp_st& a, const mpfr_t& b)
{
	return (order(a,b)<0);
}

inline int operator<(const mpfr_t& a, const mp_st& b)
{
	return (order(a,b)<0);
}

template<typename argT>
inline int operator<(const mp_st& a, const argT& b)
{
	return (order(a,b)<0);
}

template<typename argT>
inline int operator<(const argT& a, const mp_st& b)
{
	return (order(a,b)<0);
}


//
// fmax(a,b) == (a>=b) ? a : b
// fmin(a,b) == (a<b) ? a : b
// except they have a secondary criteria in case of equality based on precision
//

///////////////////////////////////////////////////////////////////////////////////////////


//****f* SAM_relations/fmax
//    NAME
//      fmax
//    SYNOPSIS
//      res = fmax(x,y) 
//    FUNCTION
//      The fmax() functions return x or y, whichever is larger.
//      
//      
//    INPUTS
//      a           - mp_st 
//      b           - mp_st
//    RESULT
//      res         - mp_st
//    SEE ALSO
//      fmin(3)
//             
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//      
//  
//    NOTES
//  
//  
//    BUGS
//  
//  
//      
//  /

inline mp_st fmax(const mp_st& a, const mp_st& b)
{
	int rank = order(a,b);
	if ( rank == 0 )
	{
		// prefer more precise parameter
		rank = a.getAccuracy() - b.getAccuracy();
	}
	if ( rank >= 0 ) {
		return a;
	}
	else {
		return b;
	}
}

template<typename argT>
inline mp_st fmax(const mp_st& a, const argT& b)
{
	int rank = order(a,b);
	if ( rank == 0 )
	{
		// prefer exact paramater
		rank = -1;
	}
	if ( rank >= 0 ) {
		return a;
	}
	else {
		return mp_st(b);
	}
}

template<typename argT>
inline mp_st fmax(const argT& a, const mp_st& b)
{
	int rank = order(a,b);
	if ( rank == 0 )
	{
		// prefer exact paramater
		rank = 1;
	}
	if ( rank >= 0 ) {
		return mp_st(a);
	}
	else {
		return b;
	}
}

///////////////////////////////////


//****f* SAM_relations/fmin
//    NAME
//      fmin
//    SYNOPSIS
//      res = fmin(x,y) 
//    FUNCTION
//      The fmin() functions return x or y, whichever is smaller.
//      
//      
//    INPUTS
//      a           - mp_st 
//      b           - mp_st
//    RESULT
//      res         - mp_st
//    SEE ALSO
//      fmax(3)
//             
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//      
//  
//    NOTES
//  
//  
//    BUGS
//  
//  
//    SEE ALSO
//      
//      
//  /

inline mp_st fmin(const mp_st& a, const mp_st& b)
{
	int rank = order(a,b);
	if ( rank == 0 )
	{
		// prefer more precise parameter (opp sign to max)
		rank = b.getAccuracy() - a.getAccuracy();
	}
	if ( rank <= 0 ) {
		return a;
	}
	else {
		return b;
	}
}

template<typename argT>
inline mp_st fmin(const mp_st& a, const argT& b)
{
	int rank = order(a,b);
	if ( rank == 0 )
	{
		// prefer exact paramater (opp sign to max
		rank = 1;
	}
	if ( rank <= 0 ) {
		return a;
	}
	else {
		return mp_st(b);
	}
}

template<typename argT>
inline mp_st fmin(const argT& a, const mp_st& b)
{
	int rank = order(a,b);
	if ( rank == 0 )
	{
		// prefer exact paramater (opp sign to max
		rank = -1;
	}
	if ( rank <= 0 ) {
		return mp_st(a);
	}
	else {
		return b;
	}
}

#endif // __SAM_RELATIONS__











