/*-----------------------------------------------------------------------
Copyright 2010-2016 S. GRAILLAT, F. JEZEQUEL
Copyright 2010 S. WANG, Y. ZHU
Copyright 2019-2020 J.E. TIDSWELL

This file is part of SAM.

    SAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    SAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
    Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with SAM. If not, see
    <http://www.gnu.org/licenses/>.
-----------------------------------------------------------------------*/
///////////////////////////////////////////////////
//
// SAM SOFTWARE
//
///////////////////////////////////////////////////
// Last Modifications
//
//
///////////////////////////////////////////////////
//****m* SAM_to/operator=
//    NAME
//      operator=
//    SYNOPSIS
//      res = v
//    FUNCTION
//      Define all the functions involving at least one argument
//      of stochastic type which overload the assignment statement "=".
//      
//    INPUTS
//      a           - an integer, a float, a double, an mpfr_t or an mp_st 
//  
//    RESULT
//      res         - an mp_st
//             
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//      
//  
//    NOTES
//  
//  
//    BUGS
//  
//  
//    SEE ALSO
//      
//      
//  /

#include <limits>
#include <cmath>

#include "SAM.h"
#include "SAM_private.h"
#include "SAM_MPFR_templates.h"

using namespace SAM;


///////////////////////////////////////////////////////////////////////////////////////////

//
// constructors
//

mp_st::mp_st() 
{
    mp_prec_t default_prec = mpfr_get_default_prec();
    mpfr_init2 (x,default_prec); 
    mpfr_init2 (y,default_prec); 
    mpfr_init2 (z,default_prec); 
#ifdef COUNT_ALLOCATIONS
	++alloc_count;
#endif
    SAM::set_mpfr (x,0,MPFR_RNDN);
    SAM::set_mpfr (y,0,MPFR_RNDN);
    SAM::set_mpfr (z,0,MPFR_RNDN);
    accuracy = (int)( default_prec * LOG ); // exact zero
}

mp_st::mp_st(const mp_st& u) // copy
{
    mp_prec_t default_prec = mpfr_get_default_prec();
    mpfr_init2 (x, default_prec);
    mpfr_init2 (y, default_prec);
    mpfr_init2 (z, default_prec);
#ifdef COUNT_ALLOCATIONS
	++alloc_count;
#endif
    SAM::set_mpfr (x, u.x, MPFR_RNDN);
    SAM::set_mpfr (y, u.y, MPFR_RNDN);
    SAM::set_mpfr (z, u.z, MPFR_RNDN);
    accuracy = u.accuracy; //Jon
    //accuracy = DIGIT_NOT_COMPUTED; //trunk
}

mp_st::mp_st(mp_st&& u) // move  //Jon
{
    this->x[0] = u.x[0];
    this->y[0] = u.y[0];
    this->z[0] = u.z[0];
#ifdef COUNT_ALLOCATIONS
	++move_count;
#endif
    this->accuracy = u.accuracy;
    u.accuracy = DIGIT_NOT_COMPUTED;
    memset(u.x,0,sizeof(mpfr_t));
    memset(u.y,0,sizeof(mpfr_t));
    memset(u.z,0,sizeof(mpfr_t));
}

//
// template trick to resolve default precision (digits)
//
template<typename T> struct mp_st_precision {
    // default is unknown ...
    static int prec(const T&) { return DIGIT_NOT_COMPUTED; } 
};
template<> struct mp_st_precision<int> {
    static int prec(const int&) { return (int)( DEFAULT_PREC * LOG ); } 
};
template<> struct mp_st_precision<unsigned int> {
    static int prec(const unsigned int&) { return (int)( DEFAULT_PREC * LOG ); } 
};
template<> struct mp_st_precision<long> {
    static int prec(const long&) { return (int)( DEFAULT_PREC * LOG ); } 
};
template<> struct mp_st_precision<unsigned long> {
    static int prec(const unsigned long&) { return (int)( DEFAULT_PREC * LOG ); } 
};
template<> struct mp_st_precision<float> {
    static int prec(const float&) { return std::numeric_limits<float>::digits10; } 
};
template<> struct mp_st_precision<double> {
    static int prec(const double&) { return std::numeric_limits<double>::digits10; }
};
template<> struct mp_st_precision<mpfr_t> {
    static int prec(const mpfr_t& u) { return (int)(mpfr_get_prec(u) * LOG); } 
};
template<> struct mp_st_precision<mpfr_srcptr> {
    static int prec(const mpfr_srcptr& u) { return (int)(mpfr_get_prec(u) * LOG); } 
};



template<typename argT> mp_st::mp_st(const argT& u)
{
    mp_prec_t default_prec = mpfr_get_default_prec();
    mpfr_init2 (x, default_prec);
    mpfr_init2 (y, default_prec);
    mpfr_init2 (z, default_prec);
#ifdef COUNT_ALLOCATIONS
	++alloc_count;
#endif

	const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	SAM::set_mpfr( x, u, xdir );
	SAM::set_mpfr( y, u, ydir );
	SAM::set_mpfr( z, u, zdir );

    const int MAX_PREC = (int)( DEFAULT_PREC * LOG );
    mpfr_prec_t p = mp_st_precision<argT>::prec(u);
    if ( p > MAX_PREC ) { p = MAX_PREC; }
    accuracy = p;
}

//
// explicit instantiation so compiler generates code to include in linker
//
template mp_st::mp_st(const int& u);
template mp_st::mp_st(const unsigned int& u);
template mp_st::mp_st(const long& u);
template mp_st::mp_st(const unsigned long& u);
template mp_st::mp_st(const float& u);
template mp_st::mp_st(const double& u);
template mp_st::mp_st(const mpfr_t& u);
template mp_st::mp_st(const mpfr_srcptr& u);



///////////////////////////////////////////////////////////////////////////////////////////

//
// assignment operators
//

mp_st& mp_st::operator=(const mp_st &v) // copy
{
    if ( this != &v )
    {
        SAM::set_mpfr(x, v.x, MPFR_RNDN);
        SAM::set_mpfr(y, v.y, MPFR_RNDN);
        SAM::set_mpfr(z, v.z, MPFR_RNDN);

        accuracy=v.accuracy;
        if ( accuracy > (int)(DEFAULT_PREC*LOG) ) { accuracy = (int)(DEFAULT_PREC*LOG); }
    }
    return *this;
}


mp_st& mp_st::operator=(mp_st &&v) // move
{
    if ( this != &v )
    {
        mpfr_swap( x, v.x );
        mpfr_swap( y, v.y );
        mpfr_swap( z, v.z );
#ifdef COUNT_ALLOCATIONS
	    ++move_count;
#endif
        std::swap( accuracy, v.accuracy );
        if ( accuracy > (int)(DEFAULT_PREC*LOG) ) { accuracy = (int)(DEFAULT_PREC*LOG); }
    }
    return *this;
}


template<typename argT> mp_st& mp_st::operator=(const argT& v)
{
    const mpfr_rnd_t xdir = (RANDOM) ? MPFR_RNDU : MPFR_RNDD;
	const bool rb2 = RANDOM;
	const mpfr_rnd_t ydir = rb2 ? MPFR_RNDU : MPFR_RNDD;
	const mpfr_rnd_t zdir = rb2 ? MPFR_RNDD : MPFR_RNDU;
	SAM::set_mpfr( x, v, xdir );
	SAM::set_mpfr( y, v, ydir );
	SAM::set_mpfr( z, v, zdir );

    const int MAX_PREC = (int)( DEFAULT_PREC * LOG );
    mpfr_prec_t p = mp_st_precision<argT>::prec(v);
    if ( p > MAX_PREC ) { p = MAX_PREC; }
    accuracy = p;
    return *this;
}

//
// explicit instantiation so compiler generates code to include in linker
//
template mp_st& mp_st::operator=<mpfr_t>(const mpfr_t&);
template mp_st& mp_st::operator=<mpfr_srcptr>(const mpfr_srcptr&);
template mp_st& mp_st::operator=<double>(const double&);
template mp_st& mp_st::operator=<float>(const float&);
template mp_st& mp_st::operator=<unsigned long>(const unsigned long&);
template mp_st& mp_st::operator=<long>(const long&);
template mp_st& mp_st::operator=<unsigned int>(const unsigned int&);
template mp_st& mp_st::operator=<int>(const int&);



///////////////////////////////////////////////////////////////////////////////////////////

//****f* SAM_to/sam_set_str
//    NAME
//     sam_set_str
//
//    SYNOPSIS
//      mp_st sam_set_str(mp_st &v, const char *s)
//    FUNCTION
//      The sam_set_str function set v to the value of the string s in base 10, 
//      rounded in the direction MPFR_RNDN. 
//    INPUTS
//      v          - a stochastic number
//      s          - a string
//    RESULT
//      v          - a stochastic number

//    SEE ALSO
//    
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//      
//  
//    NOTES
//  
//  
//    BUGS
//  
//  
//      
//      
//  /

mp_st sam_set_str(mp_st &v, const char *s)
{
	mpfr_t y;
	mpfr_init2(y,DEFAULT_PREC);
	mpfr_set_str(y,s,10,MPFR_RNDN);
		
	SAM::set_mpfr(v.x,y,MPFR_RNDN);
	SAM::set_mpfr(v.y,y,MPFR_RNDN);
	SAM::set_mpfr(v.z,y,MPFR_RNDN);
	mpfr_clear(y);
	if (_SAM_cancel_tag) {
		v.accuracy = v.nb_significant_digit(); // preemptive calculate
	} else {
		v.accuracy = DIGIT_NOT_COMPUTED;
	}
	return v;
}

/*
void old_type(mp_st &b, mp_st &a) //FJ
{
    mpfr_add(b.x,a.x,a.y, MPFR_RNDN);
    mpfr_add(b.x,b.x,a.z, MPFR_RNDN);
    mpfr_div_d(b.x,b.x,3., MPFR_RNDN); 

    SAM::set_mpfr(b.y,b.x,MPFR_RNDN);
    SAM::set_mpfr(b.z,b.x,MPFR_RNDN);
    if (_SAM_cancel_tag) {
		b.accuracy = b.nb_significant_digit(); // preemptive calculate
	} else {
		b.accuracy = DIGIT_NOT_COMPUTED;
	}
}
*/

mp_st old_type(const mp_st &a) //FJ
{   mp_st b;
    mpfr_add(b.x,a.x,a.y, MPFR_RNDN);
    mpfr_add(b.x,b.x,a.z, MPFR_RNDN);
    mpfr_div_d(b.x,b.x,3., MPFR_RNDN); 

    SAM::set_mpfr(b.y,b.x,MPFR_RNDN);
    SAM::set_mpfr(b.z,b.x,MPFR_RNDN);
    if (_SAM_cancel_tag) {
		b.accuracy = b.nb_significant_digit(); // preemptive calculate
	} else {
		b.accuracy = DIGIT_NOT_COMPUTED;
	}
    return b;
}

void SAM_prec_round(mp_st &a, int n) //FJ
{
    mpfr_prec_round(a.x,n,MPFR_RNDN);
    mpfr_prec_round(a.y,n,MPFR_RNDN);
    mpfr_prec_round(a.z,n,MPFR_RNDN);
    if (_SAM_cancel_tag) {
		a.accuracy = a.nb_significant_digit(); // preemptive calculate
	} else {
		a.accuracy = DIGIT_NOT_COMPUTED;
	}

}


//
// class methods for special case coercion
//
template<typename argT> 
mp_st mp_st::coerce(const argT& x,const argT& y, const argT& z)
{
    using namespace SAM; // 
    mp_st res;
    set_mpfr( res.x, x, MPFR_RNDN );
    set_mpfr( res.y, y, MPFR_RNDN );
    set_mpfr( res.z, z, MPFR_RNDN );
    if (_SAM_cancel_tag) {
        res.accuracy = res.nb_significant_digit(); // preemptive calculate
    } else {
        res.accuracy = DIGIT_NOT_COMPUTED;
    }
    return res;
}

//
// explicit instantiation so compiler generates code to include in linker
//
template mp_st mp_st::coerce<mpfr_t>(const mpfr_t&,const mpfr_t&, const mpfr_t&);
template mp_st mp_st::coerce<mpfr_srcptr>(const mpfr_srcptr&,const mpfr_srcptr&, const mpfr_srcptr&);
template mp_st mp_st::coerce<double>(const double&,const double&, const double&);
template mp_st mp_st::coerce<float>(const float&,const float&, const float&);
template mp_st mp_st::coerce<unsigned long>(const unsigned long&,const unsigned long&, const unsigned long&);
template mp_st mp_st::coerce<const long>(const long&,const long&, const long&);
template mp_st mp_st::coerce<unsigned int>(const unsigned int&,const unsigned int&, const unsigned int&);
template mp_st mp_st::coerce<int>(const int&,const int&, const int&);

